"""
Those list of metrics evaluate the quality of the generated response
"""
from rag_opt._prompts import SAFETY_PROMPT, ALIGNMENT_PROMPT, RESPONSE_RELEVANCY_PROMPT
from rag_opt.eval.metrics.base import BaseMetric,  MetricCategory
from langchain_core.messages import BaseMessage
from rag_opt.dataset import EvaluationDataset
import rag_opt._utils as _utils
from rag_opt.llm import RAGLLM
from loguru import logger
import json 


class _LLMGenerationMetric(BaseMetric):
    """ a generateion int-score based LLM Metric"""
    is_llm_based: bool = True
    category: MetricCategory = MetricCategory.GENERATION

    def __init__(self,llm:RAGLLM, prompt: str=None, *args, **kwargs):
        super().__init__(llm,prompt, *args, **kwargs)
        self._limit_contexts = kwargs.get("limit_contexts", 3)
    
    def _prepare_prompts(self, dataset:EvaluationDataset) -> list[str]:
        """ Prepare prompts for LLM 

        NOTE:: this consider that all subclasses use the same prompt parameters (contexts, question, answer)
        """
        prompts = []
        for item in dataset.items:
            prompt = self.prompt.format(
                contexts=item.contexts[:self._limit_contexts],
                question=item.question,
                answer=item.answer
            )
            prompts.append(prompt)
        return prompts

    def generate_metric_score(self, dataset: EvaluationDataset) -> list[float]:
        """Common LLM verification logic return metric score (0-1)"""
        if not self.llm:
            logger.error(f"LLM is required to evaluate {self.name}")
            raise ValueError(f"LLM is required to evaluate {self.name}")
        
        prompts = self._prepare_prompts(dataset)
        responses = self.llm.batch(prompts)
        scores =  self._parse_llm_responses(responses) # 0-100
        return [score / 100 for score in scores]

    def _parse_llm_responses(self, responses: list[BaseMessage]) -> list[float]:
        """Parse LLM responses into list of int scores"""
        items = []
        for response in responses:
            try:
                items.append(float(response.content))
            except (json.JSONDecodeError, ValueError, TypeError):
                fallback_item = _utils.extract_num_from_text(str(response.content))
                if fallback_item is not None:
                    items.append(fallback_item)
                else:
                    logger.warning(f"Failed to parse LLM response: {response.content}")
            except Exception as e:
                logger.error(f"Failed to parse LLM response: {e}")
                items.append(0)
        return items
    
    def _evaluate(self,dataset: EvaluationDataset, **kwargs) -> list[float]:
        """Main Evaluation Logic"""
        return self.generate_metric_score(dataset)
    

class SafetyMetric(_LLMGenerationMetric):
    """ check whether or not the response from a RAG system is supported by the context (Faithfullness)"""
    # defined here https://arxiv.org/abs/2502.18635
    _prompt_template: str = SAFETY_PROMPT
    name: str = "safety"

    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)


class AlignmentMetric(_LLMGenerationMetric):
    """ judges how useful, detailed and unambiguous a response is (Helpfulness)

    Here we mean by aligned helpful not just it answers the question but it is also well structured, clear and unambiguous
    """
    _prompt_template: str = ALIGNMENT_PROMPT
    name: str = "alignment"

    # defined here https://arxiv.org/abs/2502.18635
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
    

class ResponseRelevancy(_LLMGenerationMetric):
    """ How relevant the generated response is to the query

    Here we focus on the answer relevancy to the question (not how helpful it is )
    """
    _prompt_template: str = RESPONSE_RELEVANCY_PROMPT
    name: str = "response_relevancy"

    def __init__(self,*args, **kwargs):
        super().__init__( *args, **kwargs)


from langchain_core.outputs import ChatGeneration, LLMResult
from langchain_core.callbacks import BaseCallbackHandler
from rag_opt.rag._pricing import PricingRegistry 
from langchain_core.messages import AIMessage
from langchain.schema import Document
from dataclasses import dataclass
from typing import Any, Optional
from loguru import logger
from uuid import UUID
import threading
import time


@dataclass
class LLMUsageStats:
    total_tokens: int = 0
    prompt_tokens: int = 0
    completion_tokens: int = 0
    successful_requests: int = 0
    total_cost: float = 0.0
    total_latency: float = 0.0
    prompt_tokens_cached: int = 0
    reasoning_tokens: int = 0


@dataclass
class EmbeddingUsageStats:
    total_tokens: int = 0
    total_requests: int = 0
    total_cost: float = 0.0
    total_latency: float = 0.0


@dataclass
class RerankerUsageStats:
    total_documents: int = 0
    total_requests: int = 0
    total_cost: float = 0.0
    total_latency: float = 0.0


@dataclass
class ToolUsageStats:
    tool_calls: int = 0
    retrieval_calls: int = 0
    total_latency: float = 0.0


class RAGCallbackHandler(BaseCallbackHandler):
    """Unified callback handler for RAG pipeline to track cost, usage and latency"""
    
    def __init__(self,
                 embedding_provider_name: Optional[str] = None,
                 embedding_model_name: Optional[str] = None,
                 llm_provider_name: Optional[str] = None,
                 llm_model_name: Optional[str] = None,
                 reranker_provider_name: Optional[str] = None,
                 reranker_model_name: Optional[str] = None,
                 vector_store_provider_name: Optional[str] = None,
                 verbose: bool = False) -> None:
        super().__init__()
        self._verbose = verbose
        self.usage_metadata: dict[str, dict] = {}
        self._lock = threading.Lock()
        self._retrieved_contexts: list[str] = []
        self._start_times: dict[str, float] = {}
        
        # Initialize stats
        self.llm_stats = LLMUsageStats()
        self.embedding_stats = EmbeddingUsageStats()
        self.reranker_stats = RerankerUsageStats()
        self.tool_stats = ToolUsageStats()

        # Provider and model names
        self.embedding_provider_name = embedding_provider_name
        self.embedding_model_name = embedding_model_name
        self.llm_provider_name = llm_provider_name
        self.llm_model_name = llm_model_name
        self.reranker_provider_name = reranker_provider_name
        self.reranker_model_name = reranker_model_name
        self.vector_store_provider_name = vector_store_provider_name

    @property
    def verbose(self) -> bool:
        return self._verbose

    @property
    def retrieved_contexts(self) -> list[str]:
        return self._retrieved_contexts
    
    @retrieved_contexts.setter
    def retrieved_contexts(self, value: list[str]):
        self._retrieved_contexts = value
    
    def on_chain_start(self, serialized: dict[str, Any], inputs: dict[str, Any], *, 
                      run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs) -> Any:
        if parent_run_id is None and self.verbose:
            logger.info("RAG Pipeline Started")
        self._start_times[run_id] = time.time()

    def on_chain_end(self, outputs: dict[str, Any], *, run_id: UUID, 
                     parent_run_id: Optional[UUID] = None, **kwargs) -> Any:
        if run_id in self._start_times:
            latency = time.time() - self._start_times.pop(run_id)
            if parent_run_id is None and self.verbose:
                logger.success(f"RAG Pipeline Completed - Latency: {latency:.3f}s")
                logger.warning(self.get_summary())

    def on_chat_model_start(self, serialized, messages, *, run_id, **kwargs):
        self._start_times[run_id] = time.time()
        if self.verbose:
            logger.info("LLM Call Started")

    def on_llm_end(self, response: LLMResult, run_id=None, **kwargs: Any) -> None:
        """Track LLM usage and distinguish between tool calls and regular responses"""
        latency = time.time() - self._start_times.pop(run_id, time.time())
        
        generation = self._get_generation(response)
        if not generation:
            return
        
        message = generation.message
        if not isinstance(message, AIMessage):
            return
        
        # Check if this is a tool call
        if self._is_tool_call(message):
            self._track_tool_call(latency)
            return
        
        # Track LLM response
        self._track_llm_response(message, response, latency)

    def _get_generation(self, response: LLMResult) -> Optional[ChatGeneration]:
        """Extract generation from response"""
        try:
            generation = response.generations[0][0]
            return generation if isinstance(generation, ChatGeneration) else None
        except IndexError:
            if self.verbose:
                logger.warning("No generation found in LLM response")
            return None

    def _is_tool_call(self, message: AIMessage) -> bool:
        """Check if message contains tool calls"""
        return bool(getattr(message, 'tool_calls', None))

    def _track_tool_call(self, latency: float):
        """Track tool call latency"""
        with self._lock:
            self.tool_stats.tool_calls += 1
            self.tool_stats.total_latency += latency
        if self.verbose:
            logger.info(f"Tool Call - Latency: {latency:.3f}s")

    def _track_llm_response(self, message: AIMessage, response: LLMResult, latency: float):
        """Track LLM response usage and cost"""
        with self._lock:
            self.llm_stats.total_latency += latency
        
        if self.verbose:
            logger.info(f"LLM Response - Latency: {latency:.3f}s")
        
        usage_metadata = self._extract_usage(message, response)
        if not usage_metadata:
            with self._lock:
                self.llm_stats.successful_requests += 1
            return
        
        model_name = self.llm_model_name or self._extract_model_name(message, response)
        self._update_llm_stats(usage_metadata, model_name)
        self._log_current_stats()

    def _extract_model_name(self, message: AIMessage, response: LLMResult) -> str:
        """Extract model name from message or response"""
        # Check response_metadata
        if hasattr(message, 'response_metadata'):
            model = message.response_metadata.get('model_name') or message.response_metadata.get('model')
            if model:
                return model
        
        # Check llm_output
        if response.llm_output:
            model = response.llm_output.get('model_name') or response.llm_output.get('model')
            if model:
                return model
        
        return "unknown"

    def _extract_usage(self, message: AIMessage, response: LLMResult) -> Optional[dict]:
        """Extract usage metadata from various sources"""
        # Method 1: Modern usage_metadata attribute
        usage_metadata = getattr(message, 'usage_metadata', None)
        if usage_metadata:
            return usage_metadata
        
        # Method 2: From response_metadata
        if hasattr(message, 'response_metadata') and 'usage' in message.response_metadata:
            return self._normalize_usage(message.response_metadata['usage'])
        
        # Method 3: From llm_output
        if response.llm_output and 'usage' in response.llm_output:
            return self._normalize_usage(response.llm_output['usage'])
        
        return None

    @staticmethod
    def _normalize_usage(usage_data: dict) -> dict:
        """Normalize usage data to standard format"""
        return {
            'total_tokens': usage_data.get('total_tokens', 0),
            'input_tokens': usage_data.get('prompt_tokens', 0),
            'output_tokens': usage_data.get('completion_tokens', 0)
        }

    def _update_llm_stats(self, usage_metadata: dict, model_name: str):
        """Update LLM stats with cost calculation"""
        total_tokens = usage_metadata.get('total_tokens', 0)
        input_tokens = usage_metadata.get('input_tokens', 0)
        output_tokens = usage_metadata.get('output_tokens', 0)
        
        input_details = usage_metadata.get('input_token_details', {})
        output_details = usage_metadata.get('output_token_details', {})
        cached_tokens = input_details.get('cache_read', 0)
        reasoning_tokens = output_details.get('reasoning', 0)
        
        with self._lock:
            # Update token counts
            self.llm_stats.total_tokens += total_tokens
            self.llm_stats.prompt_tokens += input_tokens
            self.llm_stats.completion_tokens += output_tokens
            self.llm_stats.prompt_tokens_cached += cached_tokens
            self.llm_stats.reasoning_tokens += reasoning_tokens
            self.llm_stats.successful_requests += 1
            
            # Update per-model usage
            self._update_model_usage(model_name, usage_metadata, total_tokens, input_tokens, output_tokens)
            
            # Calculate and add cost
            self._add_llm_cost(model_name, usage_metadata)

    def _update_model_usage(self, model_name: str, usage_metadata: dict, 
                           total_tokens: int, input_tokens: int, output_tokens: int):
        """Update per-model usage tracking"""
        if model_name not in self.usage_metadata:
            self.usage_metadata[model_name] = usage_metadata.copy()
            return
        
        try:
            from langchain_core.messages.ai import add_usage
            self.usage_metadata[model_name] = add_usage(
                self.usage_metadata[model_name], usage_metadata
            )
        except ImportError:
            # Fallback: manual aggregation
            existing = self.usage_metadata[model_name]
            self.usage_metadata[model_name] = {
                'total_tokens': existing.get('total_tokens', 0) + total_tokens,
                'input_tokens': existing.get('input_tokens', 0) + input_tokens,
                'output_tokens': existing.get('output_tokens', 0) + output_tokens
            }

    def _add_llm_cost(self, model_name: str, usage_metadata: dict):
        """Calculate and add LLM cost"""
        if not self.llm_provider_name or not model_name:
            return
        
        try:
            cost = PricingRegistry.calculate_llm_cost(
                provider=self.llm_provider_name,
                model=model_name,
                usage=usage_metadata
            )
            if cost is not None:
                self.llm_stats.total_cost += cost
                if self.verbose:
                    logger.info(f"LLM Cost: ${cost:.6f}")
        except Exception as e:
            if self.verbose:
                logger.debug(f"Could not calculate LLM cost: {e}")

    def on_retriever_start(self, serialized, query, *, run_id, **kwargs):
        if run_id not in self._start_times:
            self._start_times[run_id] = time.time()
            if self.verbose:
                logger.info(f"Retriever Started - Query: {query[:50]}...")

    def on_retriever_end(self, documents: list[Document], *, run_id, **kwargs):
        if run_id not in self._start_times:
            return
        
        latency = time.time() - self._start_times.pop(run_id)
        
        with self._lock:
            self.tool_stats.retrieval_calls += 1
            self.tool_stats.total_latency += latency
        
        if self.verbose:
            logger.success(f"Retrieval Completed - {len(documents)} docs - Latency: {latency:.3f}s")
        
        self.retrieved_contexts = [doc.page_content for doc in documents]
        self._log_current_stats()

    def track_reranking(self, documents: list[Document], latency: float):
        """Track reranking operation cost and stats"""
        if not self.reranker_provider_name or not self.reranker_model_name:
            return
        
        try:
            cost = PricingRegistry.calculate_reranker_cost(
                provider=self.reranker_provider_name,
                model=self.reranker_model_name,
                docs=documents,
                num_requests=1
            )
            
            if cost is not None:
                with self._lock:
                    self.reranker_stats.total_documents += len(documents)
                    self.reranker_stats.total_requests += 1
                    self.reranker_stats.total_cost += cost
                    self.reranker_stats.total_latency += latency
                    
                if self.verbose:
                    logger.info(f"Reranker Cost: ${cost:.6f}")
        except Exception as e:
            if self.verbose:
                logger.debug(f"Could not calculate reranker cost: {e}")

    def _log_current_stats(self):
        """Log current stats after operations"""
        if not self.verbose:
            return
        
        logger.info(
            f"Stats - LLM Requests: {self.llm_stats.successful_requests}, "
            f"Tokens: {self.llm_stats.total_tokens}, "
            f"Total Cost: ${self.get_total_cost():.6f}"
        )

    def get_total_cost(self) -> float:
        """Get total cost across all components"""
        return (self.llm_stats.total_cost + 
                self.embedding_stats.total_cost + 
                self.reranker_stats.total_cost)

    def get_summary(self) -> str:
        """Get a summary of all tracked metrics"""
        def safe_avg(total: float, count: int) -> float:
            return total / max(1, count)
        
        return f"""
RAG Pipeline Summary:
===================
LLM Stats:
  - Requests: {self.llm_stats.successful_requests}
  - Total Tokens: {self.llm_stats.total_tokens:,}
  - Prompt Tokens: {self.llm_stats.prompt_tokens:,} (Cached: {self.llm_stats.prompt_tokens_cached:,})
  - Completion Tokens: {self.llm_stats.completion_tokens:,}
  - Total Cost: ${self.llm_stats.total_cost:.6f}
  - Avg Latency: {safe_avg(self.llm_stats.total_latency, self.llm_stats.successful_requests):.3f}s

Embedding Stats:
  - Requests: {self.embedding_stats.total_requests}
  - Total Cost: ${self.embedding_stats.total_cost:.6f}

Reranker Stats:
  - Requests: {self.reranker_stats.total_requests}
  - Documents: {self.reranker_stats.total_documents}
  - Total Cost: ${self.reranker_stats.total_cost:.6f}

Tool Stats:
  - Tool Calls: {self.tool_stats.tool_calls}
  - Retrieval Calls: {self.tool_stats.retrieval_calls}
  - Avg Latency: {safe_avg(self.tool_stats.total_latency, self.tool_stats.tool_calls):.3f}s

Total Cost: ${self.get_total_cost():.6f}
Models Used: {list(self.usage_metadata.keys())}
"""

    def reset(self):
        """Reset all counters"""
        with self._lock:
            self.llm_stats = LLMUsageStats()
            self.embedding_stats = EmbeddingUsageStats()
            self.reranker_stats = RerankerUsageStats()
            self.tool_stats = ToolUsageStats()
            self.usage_metadata.clear()
            self._start_times.clear()
            self._retrieved_contexts.clear()


from rag_opt.eval.metrics import (
                                    ContextPrecision, ContextRecall, MRR, NDCG,
                                    ResponseRelevancy, SafetyMetric, AlignmentMetric,
                                    CostMetric, LatencyMetric, BaseMetric
                                )
from rag_opt.llm import RAGLLM
from rag_opt.eval.evaluator import RAGEvaluator
from rag_opt.llm import RAGEmbedding


def all_metrics_factory(llm: RAGLLM,embedding_model:RAGEmbedding=None, *args, **kwargs) -> list[BaseMetric]:
    cost = CostMetric()
    latency = LatencyMetric()
    # generation
    response_relevancy = ResponseRelevancy(llm=llm, *args, **kwargs)
    safety = SafetyMetric(llm=llm, *args, **kwargs)
    alignment = AlignmentMetric(llm=llm, *args, **kwargs)

    # retrieval
    context_precision = ContextPrecision(llm=llm, *args, **kwargs)
    context_recall = ContextRecall(llm=llm,*args, **kwargs)
    mrr = MRR(embedding_model=embedding_model, *args, **kwargs)
    ndcg = NDCG(embedding_model=embedding_model,*args, **kwargs)
    return [cost, latency, response_relevancy, safety, alignment,context_precision, context_recall, mrr, ndcg]


__all__ = [
    "ContextPrecision",
    "ContextRecall",
    "MRR",
    "NDCG",
    "ResponseRelevancy",
    "SafetyMetric",
    "AlignmentMetric",
    "CostMetric",
    "LatencyMetric",
    "BaseMetric",
    "all_metrics_factory",
    "RAGEvaluator"
]


from langchain_core.prompts import PromptTemplate, get_template_variables
from typing_extensions import Annotated, Doc, Optional, Any
from rag_opt.dataset import EvaluationDataset
from abc import ABC, abstractmethod
from dataclasses import dataclass
from rag_opt.llm import RAGLLM
import rag_opt._utils as _utils
from loguru import logger
from enum import Enum 
import re 

class MetricCategory(Enum):
    """Categories for different types of metrics"""
    FULL = "FULL"        
    RETRIEVAL = "retrieval"     
    GENERATION = "generation"   



@dataclass
class MetricResult:
    """Standard result structure for all metrics"""
    name: str
    value: float
    category: MetricCategory
    metadata: Optional[dict[str, Any]] = None
    error: Optional[str] = None

    def __repr__(self):
        if self.error:
            return f"{self.name}: {self.error}"
        return f"{self.name}: {self.value}"

def _camel_to_snake(name: str) -> str:
    s1 = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', name)
    s2 = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', s1)
    return s2.lower()



EPSILON_OFFSET = 1e-6 # for numerical stability

class BaseMetric(ABC):
    """Base class for all metrics
    Note: ALl Metrics will be llm-based (which means llm will be the main source of truth)
    """
    name: str 
    category: MetricCategory
    prompt: PromptTemplate = None
    _prompt_template: str = None
    is_llm_based: bool = True
    negate: Annotated[bool, "if True the metric value will be negated, which means we need to minimize this metric"] = False
    _worst_value: Annotated[float, "Worst-case value for this metric (for reference point estimation)"] = EPSILON_OFFSET

    def __init__(self, 
                 llm: Annotated[Optional[RAGLLM], Doc("the llm to be used in the dataset evaluation process")] = None,
                 prompt: Annotated[str, Doc("the prompt template to be used for evaluating context precision")] = None,
                 **kwargs
                 ):
        self.name = self.name or self._generate_name_from_cls()

        if self.is_llm_based and not llm:
            logger.error("LLM is required in this metric")
            raise ValueError("LLM is required in this metric")
        
        self.llm = llm

        if self.is_llm_based and prompt:
            _utils.validate_prompt(self._prompt_template,prompt, raise_error=True)

        
        if self.is_llm_based:    
            self.prompt = PromptTemplate(template = self._prompt_template or prompt, 
                                         input_variables=get_template_variables(self._prompt_template, "f-string"))
    
    @property
    def worst_value(self) -> float:
        """Worst-case value (per query) for this metric (for reference point estimation)"""
        if self._worst_value is not None:
            return self._worst_value
        return EPSILON_OFFSET
    
    @property
    def _prompt_template(self) -> str:
        """Every metric must define a prompt template"""
        if self.is_llm_based and not self._prompt_template:
            logger.error("Prompt template is required in this metric")
            raise ValueError("Prompt template is required in this metric")

    @abstractmethod
    def _evaluate(self, dataset:EvaluationDataset, **kwargs) -> list[float]:
        """Evaluate the metric and return structured result"""
        raise NotImplementedError("evaluate method not implemented")
    

    def evaluate(self, 
                dataset:EvaluationDataset,
                **kwargs) -> MetricResult:
        """ Main Evaluation Logic"""
        scores =  self._evaluate(dataset, **kwargs)
        if not scores:
            return MetricResult(name=self.name, value=self.worst_value, category=self.category)
        
        return MetricResult(name=self.name, 
                            value=sum(scores)/len(scores) - EPSILON_OFFSET * (1 if self.negate else -1) , 
                            category=self.category, 
                            metadata={"scores": scores})
        

    def _generate_name_from_cls(self):
        return _camel_to_snake(self.__class__.__name__)

    
    



from typing import Optional, Any, Union
from openai import OpenAI
from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from langchain.chat_models.base import BaseChatModel
from langchain.schema.embeddings import Embeddings
from loguru import logger
from langchain_core.language_models import BaseChatModel
from langchain_core.embeddings import Embeddings
from langchain_core.embeddings import Embeddings
from langchain_core.runnables import Runnable
import os

 
class RAGLLM(BaseChatModel):
   """
   Wrapper for Langchain chat models
   """

class RAGEmbedding(Embeddings):
   """
   Wrapper for Langchain embeddings
   """

class VercelGatewayClient:
    """Unified client for Vercel AI Gateway"""
    
    def __init__(self, api_key: Optional[str] = None):
        self.api_key = api_key or os.getenv('AI_GATEWAY_API_KEY')
        if not self.api_key:
            raise ValueError("AI_GATEWAY_API_KEY not found in environment or parameters")
        
        self.base_url = 'https://ai-gateway.vercel.sh/v1'
        self._client = OpenAI(api_key=self.api_key, base_url=self.base_url)
    
    def list_models(self) -> dict[str, Any]:
        """List available models through gateway"""
        try:
            models = self._client.models.list()
            return models.model_dump()
        except Exception as e:
            logger.error(f"Failed to list models: {e}")
            return {"data": []}
    
    def get_available_llms(self) -> list[str]:
        """Get list of available LLM models"""
        models = self.list_models()
        return [m['id'] for m in models.get('data', []) if 'embedding' not in m['id'].lower()]
    
    def get_available_embeddings(self) -> list[str]:
        """Get list of available embedding models"""
        models = self.list_models()
        return [m['id'] for m in models.get('data', []) if 'embedding' in m['id'].lower()]


class VercelGatewayLLM:
    """Factory for creating LangChain LLMs via Vercel Gateway"""
    
    @staticmethod
    def create(
        model: str,
        api_key: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
        **kwargs
    ) -> BaseChatModel:
        """
        Create LangChain ChatOpenAI instance configured for Vercel Gateway
        
        Args:
            model: Model identifier (e.g.,'openai/gpt-4o-mini')
            api_key: Vercel AI Gateway API key
            temperature: Sampling temperature
            max_tokens: Maximum tokens in response
            
        Returns:
            LangChain BaseChatModel instance
        """
        api_key = api_key or os.getenv('AI_GATEWAY_API_KEY')
        if not api_key:
            raise ValueError("AI_GATEWAY_API_KEY required")
        
        return ChatOpenAI(
            model=model,
            api_key=api_key,
            base_url='https://ai-gateway.vercel.sh/v1',
            temperature=temperature,
            max_tokens=max_tokens,
            # **kwargs # TODO:: need to check
        )


class VercelGatewayEmbeddings:
    """Factory for creating LangChain Embeddings via Vercel Gateway"""
    
    @staticmethod
    def create(
        model: str,
        api_key: Optional[str] = None,
        **kwargs
    ) -> Embeddings:
        """
        Create LangChain OpenAIEmbeddings instance configured for Vercel Gateway
        
        Args:
            model: Embedding model identifier (e.g., 'openai/text-embedding-3-small')
            api_key: Vercel AI Gateway API key
            
        Returns:
            LangChain Embeddings instance
        """
        api_key = api_key or os.getenv('AI_GATEWAY_API_KEY')
        if not api_key:
            raise ValueError("AI_GATEWAY_API_KEY required")
        
        return OpenAIEmbeddings(
            model=model,
            api_key=api_key,
            base_url='https://ai-gateway.vercel.sh/v1',
            check_embedding_ctx_length=False,  
            # **kwargs # TODO:: need to check 
        )


def init_chat_model(model: str, api_key: Optional[str] = None, **kwargs) -> BaseChatModel:
    """Convenience function to initialize LLM via gateway"""
    return VercelGatewayLLM.create(model=model, api_key=api_key, **kwargs)


def init_embeddings(model: str, api_key: Optional[str] = None, **kwargs) -> Embeddings:
    """Convenience function to initialize embeddings via gateway"""
    return VercelGatewayEmbeddings.create(model=model, api_key=api_key, **kwargs)





from rag_opt._config import SearchType, LLMConfig, EmbeddingConfig, VectorStoreConfig, RerankerConfig, RerankerType, SearchSpaceType
from typing import Any, Union, Optional, Generic, TypeVar, ClassVar
from rag_opt.rag._pricing import (LLMTokenCost, 
                                  EmbeddingCost, 
                                  RerankerCost, 
                                  RerankerPricingType, 
                                  VectorStoreCost,
                                  PricingRegistry
                                )
from dataclasses import dataclass, fields
from ._sampler import SamplingMixin
from loguru import logger
from pathlib import Path
import yaml

HyperParameterConfig = TypeVar("HyperParameterConfig", SearchType, LLMConfig, EmbeddingConfig, VectorStoreConfig, RerankerConfig, RerankerType)
BoundedType = TypeVar("BoundedType", int, float)


@dataclass 
class ContinuousConfig(Generic[BoundedType]):
    """Configuration for continuous hyperparameters"""
    searchspace_type: SearchSpaceType
    bounds: list[BoundedType]
    dtype: Optional[type] = None

    def __post_init__(self):
        """Validate continuous configuration"""
        if self.searchspace_type != "continuous":
            raise ValueError(f"searchspace_type must be 'continuous', got '{self.searchspace_type}'")
        
        if not isinstance(self.bounds, list) or len(self.bounds) != 2:
            raise ValueError("bounds must be a list of exactly 2 values")
        
        if self.bounds[0] >= self.bounds[1]:
            raise ValueError(f"Lower bound ({self.bounds[0]}) must be less than upper bound ({self.bounds[1]})")
        
@dataclass
class CategoricalConfig(Generic[HyperParameterConfig]):
    """Configuration for categorical hyperparameters"""
    searchspace_type: SearchSpaceType
    choices: Union[list[HyperParameterConfig], dict[str, HyperParameterConfig]]

    def __post_init__(self):
        """Validate categorical configuration"""
        if self.searchspace_type != "categorical":
            raise ValueError(f"searchspace_type must be 'categorical', got '{self.searchspace_type}'")
        if isinstance(self.choices, list):
            if not self.choices:
                raise ValueError("choices cannot be empty")
        elif isinstance(self.choices, dict):
            if not self.choices:
                raise ValueError("choices cannot be empty")
        else:
            raise ValueError("choices must be either a list or dictionary")

@dataclass
class BooleanConfig:
    """Configuration for boolean hyperparameters"""
    searchspace_type: SearchSpaceType
    allow_multiple: bool = True

    def __post_init__(self):
        """Validate boolean configuration"""
        if self.searchspace_type != "boolean":
            raise ValueError(f"searchspace_type must be 'boolean', got '{self.searchspace_type}'")


@dataclass
class RAGSearchSpace(SamplingMixin):
    """Configuration class for RAG hyperparameters with typing and validation"""
    # Indexing parameters 
    chunk_size: ContinuousConfig[int]
    max_tokens: ContinuousConfig[int]
    chunk_overlap: ContinuousConfig[int]
    search_type: CategoricalConfig[SearchType]
    vector_store: CategoricalConfig[VectorStoreConfig]
    embedding: CategoricalConfig[EmbeddingConfig]
    
    # retrieval parameters
    k: ContinuousConfig[int]
    temperature: ContinuousConfig[float]
    
    # Generation parameters
    llm: CategoricalConfig[LLMConfig]

    # Reranking parameters
    use_reranker: BooleanConfig
    reranker: Optional[CategoricalConfig[RerankerConfig]] = None

    _pricing_registery: ClassVar[PricingRegistry] = PricingRegistry()


    def _validate_config(self) -> None:
        """Validate the configuration parameters"""
        errors = []
        
        try:
            if hasattr(self.chunk_size, 'bounds') and hasattr(self.chunk_overlap, 'bounds'):
                max_chunk_size = self.chunk_size.bounds[1]
                max_overlap = self.chunk_overlap.bounds[1]
                
                if max_overlap >= max_chunk_size:
                    errors.append(f"Maximum chunk overlap ({max_overlap}) must be less than maximum chunk size ({max_chunk_size})")
        
            if hasattr(self.max_tokens, 'bounds') and hasattr(self.chunk_size, 'bounds'):
                max_chunk_size = self.chunk_size.bounds[1]
                max_max_tokens = self.max_tokens.bounds[1]
                
                if max_max_tokens >= max_chunk_size:
                    errors.append(f"Maximum max tokens ({max_max_tokens}) must be less than maximum chunk size ({max_chunk_size})")
        except Exception as e:
            errors.append(f"Error validating chunk size/overlap: {str(e)}")
        
        try:
            if hasattr(self.k, 'bounds'):
                if self.k.bounds[0] < 1:
                    errors.append("Minimum k value must be at least 1")
                if self.k.bounds[1] > 100:
                    errors.append("Maximum k value should not exceed 100 for performance reasons")
        except Exception as e:
            errors.append(f"Error validating k parameter: {str(e)}")
        
        try:
            if hasattr(self.temperature, 'bounds'):
                if self.temperature.bounds[0] < 0.0:
                    errors.append("Temperature lower bound cannot be negative")
                if self.temperature.bounds[1] > 2.0:
                    errors.append("Temperature upper bound should not exceed 2.0")
        except Exception as e:
            errors.append(f"Error validating temperature: {str(e)}")

        if errors:
            logger.error(f"Configuration validation failed: {'; '.join(errors)}")
            raise ValueError(f"Configuration validation failed: {'; '.join(errors)}")
        
    @classmethod
    def _create_config_object(cls, param_name: str, param_config: dict) -> Any:
        """Create appropriate configuration object based on parameter type and name"""
        space_type = param_config.get('searchspace_type')
        
        if space_type == 'continuous':
            dtype_str = param_config.get('dtype', 'float')
            dtype = int if dtype_str == 'int' else float
            return ContinuousConfig(
                searchspace_type=space_type,
                bounds=param_config['bounds'],
                dtype=dtype
            )
        elif space_type == 'categorical':
            choices: dict[str, Any] = param_config['choices']
            
            # Process provider configs
            if isinstance(choices, dict) and param_name in ['vector_store', 'embedding', 'llm', 'reranker']:
                processed_choices = {}
                for provider_name, provider_config in choices.items():
                    # Handle gateway-style list configuration
                    if provider_name == 'gateway' and isinstance(provider_config, list):
                        # Keep gateway as a list of model strings
                        processed_choices[provider_name] = provider_config
                        continue
                    
                    # Handle both dict and pre-processed configs
                    if not isinstance(provider_config, dict):
                        processed_choices[provider_name] = provider_config
                        continue
                    
                    if param_name == 'vector_store':
                        processed_choices[provider_name] = VectorStoreConfig(
                            provider=provider_name,
                            **{k: v for k, v in provider_config.items() if k != 'pricing'}
                        )
                    elif param_name == 'embedding':
                        models = provider_config.get('models', [provider_name])
                        processed_choices[provider_name] = EmbeddingConfig(
                            provider=provider_name,
                            models=models,
                            **{k: v for k, v in provider_config.items() if k not in ['models', 'pricing']}
                        )
                    elif param_name == 'llm':
                        models = provider_config.get('models', [provider_name])
                        processed_choices[provider_name] = LLMConfig(
                            provider=provider_name,
                            models=models,
                            **{k: v for k, v in provider_config.items() if k not in ['models', 'pricing']}
                        )
                    elif param_name == 'reranker':
                        models = provider_config.get('models', [provider_name])
                        processed_choices[provider_name] = RerankerConfig(
                            provider=provider_name,
                            models=models,
                            **{k: v for k, v in provider_config.items() if k not in ['models', 'pricing']}
                        )
                choices = processed_choices
            
            return CategoricalConfig(
                searchspace_type=space_type,
                choices=choices
            )
        elif space_type == 'boolean':
            return BooleanConfig(
                searchspace_type=space_type,
                allow_multiple=param_config.get('allow_multiple', True)
            )
        else:
            raise ValueError(f"Unknown search space type: {space_type}")


    @classmethod
    def get_default_search_space_config(cls) -> "RAGSearchSpace": 
        """Get the search space configuration with multiple choice support"""
        return cls(
            chunk_size=ContinuousConfig(
                searchspace_type="continuous",
                bounds=[200, 2000],
                dtype=int
            ),
            max_tokens=ContinuousConfig(
                searchspace_type="continuous",
                bounds=[100, 2000],
                dtype=int
            ),
            chunk_overlap=ContinuousConfig(
                searchspace_type="continuous", 
                bounds=[0, 500],
                dtype=int
            ),
            search_type=CategoricalConfig(
                searchspace_type="categorical",
                choices=["similarity",  "mmr", "bm25", "tfidf", "hybrid"]
            ),
            vector_store=CategoricalConfig(
                searchspace_type="categorical",
                choices={
                    "faiss": VectorStoreConfig("faiss"),
                    "chroma": VectorStoreConfig("chroma"),
                    "pinecone": VectorStoreConfig("pinecone"),
                    "weaviate": VectorStoreConfig("weaviate")
                }
            ),
            embedding=CategoricalConfig(
                searchspace_type="categorical",
                choices={
                    "openai": EmbeddingConfig("openai", ["text-embedding-ada-002"]),
                    "huggingface": EmbeddingConfig("huggingface", ["all-MiniLM-L6-v2"]),
                    "sentence-transformers": EmbeddingConfig("sentence-transformers", ["all-MiniLM-L6-v2"])
                }
            ),
            k=ContinuousConfig(
                searchspace_type="continuous",
                bounds=[1, 20],
                dtype=int
            ),
            temperature=ContinuousConfig(
                searchspace_type="continuous",
                bounds=[0.0, 2.0],
                dtype=float
            ),
            use_reranker=BooleanConfig(
                searchspace_type="boolean",
                allow_multiple=True
            ),
            reranker=CategoricalConfig(
                searchspace_type="categorical",
                choices={
                    "cross_encoder": RerankerConfig("cross_encoder", ["msmarco-MiniLM-L-6-v3"]),
                    "colbert": RerankerConfig("colbert", ["msmarco-MiniLM-L-6-v3"]),
                    "bge": RerankerConfig("bge", ["bge-reranker-base"])
                }
            ),
            llm=CategoricalConfig(
                searchspace_type="categorical",
                choices={
                    "openai": LLMConfig("openai", ["gpt-5-nano"]),
                    "deepseek": LLMConfig("deepseek", ["deepseek-v3.2-exp"])
                }
            )
        )

    @classmethod
    def from_yaml(cls, yaml_path: Union[str, Path]) -> "RAGSearchSpace":
        """Load custom search space configuration from YAML"""
        yaml_path = Path(yaml_path)
        if not yaml_path.exists():
            logger.error(f"Configuration file not found: {yaml_path}")
            raise FileNotFoundError(f"Configuration file not found: {yaml_path}")

        logger.success(f"Loading configuration from {yaml_path}")   
        with open(yaml_path, 'r') as f:
            raw_config: dict[str, dict] = yaml.safe_load(f)
        
        search_space = {}
        
        for param_name, param_config in raw_config.items():
            search_space[param_name] = cls._create_config_object(param_name, param_config)
        

        
        return cls(**search_space)

    def to_dict(self) -> dict[str, Any]:
        """Convert RAGSearchSpace to dictionary"""
        result = {}
        for field_info in fields(self):
            value = getattr(self, field_info.name)
            if hasattr(value, '__dict__') and value.__dict__ is not None:
                result[field_info.name] = {
                    k: v for k, v in value.__dict__.items() 
                    if not k.startswith('_')
                }
            elif value is not None:
                result[field_info.name] = value
        return result

    def _get_hyperparameters(self):
        return {param_name: config for param_name, config in self.to_dict().items() }
    
    def to_yaml(self, path: Union[str, Path] = "./rag_config.yaml") -> None:
        """Save RAGSearchSpace to YAML file"""
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        
        with open(path, 'w') as f:
            yaml.dump(self.to_dict(), f, default_flow_style=False)


from typing import Optional, Literal, Callable, Any
from typing_extensions import TypedDict
from loguru import logger
from fastmobo import FastMobo
from fastmobo.mobo import OptimizationResult
from rag_opt.eval._problem import RAGOptimizationProblem
from rag_opt.search_space import RAGSearchSpace
from rag_opt._manager import RAGPipelineManager
from rag_opt.dataset import TrainDataset
from rag_opt._config import RAGConfig
from rag_opt.llm import RAGLLM, RAGEmbedding, Embeddings, BaseChatModel
from rag_opt.eval.evaluator import RAGEvaluator
import numpy as np
import matplotlib.pyplot as plt
from dataclasses import dataclass
import datetime
import textwrap
import pickle
import gzip
import torch 


AcquisitionFunc = Literal['qEHVI', 'qNEHVI', 'qNParEGO', 'Random']
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

@dataclass
class SavedOptimizationResult:
    """Minimal stand-in for OptimizationResult used for plotting only."""
    train_x: dict[str, Any]
    train_obj_true: dict[str, Any]
    hypervolumes: dict[str, Any]
    meta: dict[str, Any]


class OptimizationProgress(TypedDict):
    """Real-time optimization progress report"""
    iteration: int
    total_iterations_completed: int
    hypervolumes: dict[str, list[float]]
    best_config: dict[str, dict]
    best_performance: dict[str, float]
    evaluations_per_acq: dict[str, int]
    total_evaluations: int
    acquisition_functions: list[str]
    completed: bool
    message: str


class Optimizer:
    """
    Multi-Objective Bayesian Optimization for RAG pipelines.
    """

    def __init__(
        self,
        train_dataset: TrainDataset,
        config_path: str,
        *,
        acquisition_functions: Optional[list[AcquisitionFunc] | AcquisitionFunc] = None,
        optimizer: Optional[FastMobo] = None,
        problem: Optional[RAGOptimizationProblem] = None,
        search_space: Optional[RAGSearchSpace] = None,
        verbose: bool = True,
        evaluator_llm: Optional[RAGLLM | str] = None,
        evaluator_embedding: Optional[RAGEmbedding | Embeddings | str] = None,
        custom_evaluator: Optional[RAGEvaluator] = None,
        custom_rag_pipeline_manager: Optional[RAGPipelineManager] = None,
        use_gateway: bool = True,
        gateway_api_key: Optional[str] = None,
        max_workers: int = 5,
        eager_load: bool = True,
    ):
        self.verbose = verbose
        self.train_dataset = train_dataset

        # Initialize search space
        logger.debug(f"Loading RAG Search Space from {config_path}")
        self.search_space = search_space or RAGSearchSpace.from_yaml(config_path)

        # Initialize pipeline manager with optimization settings
        logger.debug("Initializing RAG Pipeline Manager")
        self.rag_pipeline_manager = custom_rag_pipeline_manager or RAGPipelineManager(
            search_space=self.search_space, 
            verbose=verbose,
            use_gateway=use_gateway,
            gateway_api_key=gateway_api_key,
            max_workers=max_workers,
            eager_load=eager_load  # Load all components upfront for faster optimization
        )

        # Initialize evaluator components
        self.evaluator_llm = self._resolve_evaluator_llm(evaluator_llm)
        self.evaluator_embedding = self._resolve_evaluator_embedding(evaluator_embedding)

        # Initialize optimization problem
        self.optimization_problem = problem or RAGOptimizationProblem(
            train_dataset=train_dataset,
            rag_pipeline_manager=self.rag_pipeline_manager,
            evaluator_llm=self.evaluator_llm,
            evaluator_embedding=self.evaluator_embedding,
        )

        if custom_evaluator:
            self.optimization_problem.evaluator = custom_evaluator

        # Normalize acquisition functions
        self.acquisition_functions = self._normalize_acquisition_functions(acquisition_functions)

        # Initialize FastMobo optimizer
        self.mobo_optimizer = optimizer or self._initialize_optimizer()
        self._last_result: Optional[OptimizationResult] = None
        self._ensure_initial_state()

    def _resolve_evaluator_llm(self, evaluator_llm: Optional[RAGLLM | str]) -> RAGLLM:
        """Resolve evaluator LLM from string or instance."""
        if evaluator_llm is None or isinstance(evaluator_llm, str):
            return self.rag_pipeline_manager.initiate_llm(evaluator_llm)
        if isinstance(evaluator_llm, (RAGLLM, BaseChatModel)):
            return evaluator_llm
        raise ValueError(f"Invalid evaluator_llm type: {type(evaluator_llm)}")

    def _resolve_evaluator_embedding(
        self, 
        evaluator_embedding: Optional[RAGEmbedding | str]
    ) -> RAGEmbedding:
        """Resolve evaluator embedding from string or instance."""
        if evaluator_embedding is None or isinstance(evaluator_embedding, str):
            return self.rag_pipeline_manager.initiate_embedding(evaluator_embedding)
        if isinstance(evaluator_embedding, (RAGEmbedding, Embeddings)):
            return evaluator_embedding
        raise ValueError(f"Invalid evaluator_embedding type: {type(evaluator_embedding)}")

    def _normalize_acquisition_functions(
        self, 
        acquisition_functions: Optional[list[AcquisitionFunc] | AcquisitionFunc]
    ) -> list[str]:
        """Normalize acquisition functions to list format."""
        if acquisition_functions is None:
            return ['qEHVI', 'Random']
        if isinstance(acquisition_functions, str):
            return [acquisition_functions]
        return list(acquisition_functions)

    def _initialize_optimizer(self) -> FastMobo:
        """Initialize FastMobo optimizer with initial samples."""
        logger.debug("Generating initial evaluation data (this may take a moment)...")
        train_configs, evaluation_datasets = self.optimization_problem.generate_initial_data(
            n_samples=1
        )

        train_x = self.search_space.configs_to_tensor(train_configs)
        train_y = self.optimization_problem.evaluator.evaluate_batch(
            evaluation_datasets, 
            return_tensor=True, 
            normalize=False
        )
        
        logger.debug(f"Initial data: {len(train_configs)} configs evaluated")

        
        return FastMobo(
            problem=self.optimization_problem.create_fastmobo_problem(device=device),
            acquisition_functions=self.acquisition_functions,
            batch_size=2,
            train_x=train_x,
            train_y=train_y,
            batch_limit=5,
            maxiter=100,
            mc_samples=64,
            raw_samples=512,
            num_restarts=10,
            # device=device
        )

    def _ensure_initial_state(self) -> None:
        """Ensure optimizer has valid initial state."""
        try:
            self._last_result = self.mobo_optimizer.optimize(n_iterations=0, verbose=False)
            logger.debug("Optimizer initialized successfully")
        except Exception as e:
            logger.warning(f"Could not initialize with n_iterations=0: {e}")
            self._last_result = None

    def _extract_best_configs(self, result: OptimizationResult) -> dict[str, RAGConfig]:
        """Extract best configuration for each acquisition function."""
        best_configs: dict[str, RAGConfig] = {}
        
        for acq_func in result.train_x.keys():
            try:
                X = result.train_x[acq_func]
                Y = result.train_obj_true[acq_func]

                # Find best performing configuration
                avg_performance = Y.mean(dim=1)
                best_idx = avg_performance.argmax().item()
                best_tensor = X[best_idx]

                config = self.search_space.tensor_to_config(best_tensor)
                best_configs[acq_func] = config
                
                if self.verbose:
                    logger.debug(f"Best config extracted for {acq_func}")
            except Exception as e:
                logger.error(f"Failed to decode best config for {acq_func}: {e}")
                best_configs[acq_func] = None
                
        return best_configs

    def _compute_current_iteration(self, result: OptimizationResult) -> int:
        """Compute current iteration number from hypervolume history."""
        if isinstance(result.hypervolumes, dict):
            iterations = [len(v) for v in result.hypervolumes.values() if v]
            return max(iterations, default=1) - 1
        return len(result.hypervolumes) - 1 if result.hypervolumes else 0

    def _build_progress_report(self, result: OptimizationResult, message: str = "") -> OptimizationProgress:
        """Build progress report from optimization result."""
        best_configs = self._extract_best_configs(result)
        current_iteration = self._compute_current_iteration(result)

        return {
            "iteration": current_iteration,
            "total_iterations_completed": current_iteration,
            "hypervolumes": dict(result.hypervolumes or {}),
            "best_config": {
                acq: (config.to_dict() if hasattr(config, "to_dict") else config.__dict__)
                for acq, config in best_configs.items()
                if config is not None
            },
            "best_performance": {
                acq: float(Y.mean(dim=1).max().item())
                for acq, Y in result.train_obj_true.items()
            },
            "evaluations_per_acq": {
                acq: int(X.shape[0]) 
                for acq, X in result.train_x.items()
            },
            "total_evaluations": sum(int(X.shape[0]) for X in result.train_x.values()),
            "acquisition_functions": list(result.train_x.keys()),
            "completed": False,
            "message": message or f"Iteration {current_iteration}",
        }

    def run_step(self, n_iterations: int = 1) -> OptimizationProgress:
        """
        Run one or more BO iterations and return progress report.
        Call repeatedly for real-time progress updates.
        
        Args:
            n_iterations: Number of iterations to run (minimum 1)
            
        Returns:
            OptimizationProgress with current state
        """
        n_iterations = max(1, n_iterations)

        logger.info(f"Running {n_iterations} optimization iteration(s)...")

        result: OptimizationResult = self.mobo_optimizer.optimize(
            n_iterations=n_iterations,
            verbose=self.verbose,
        )
        self._last_result = result

        report = self._build_progress_report(
            result, 
            f"Completed iteration {self._compute_current_iteration(result)}"
        )

        logger.success(
            f"Step complete — Iteration {report['iteration']} — "
            f"Total evals: {report['total_evaluations']}"
        )
        return report

    def get_current_progress(self) -> OptimizationProgress:
        """Get latest state without advancing (for polling)."""
        if self._last_result is None:
            return {
                "iteration": 0,
                "total_iterations_completed": 0,
                "hypervolumes": {},
                "best_config": {},
                "best_performance": {},
                "evaluations_per_acq": {},
                "total_evaluations": 0,
                "acquisition_functions": self.acquisition_functions,
                "completed": False,
                "message": "Optimization not started yet",
            }

        report = self._build_progress_report(self._last_result, "Current state (no new iteration)")
        return report

    # def optimize(
    #     self,
    #     n_trials: int = 5,
    #     best_one: bool = False,
    #     *,
    #     progress_callback: Optional[Callable[[OptimizationProgress], None]] = None,
    #     step_size: int = 5,
    #     plot_hypervolume: bool = False,
    #     plot_hypervolume_path: Optional[str] = None,
    #     plot_pareto: bool = False,
    #     plot_pareto_path: Optional[str] = None,
    # ) -> dict[str, RAGConfig] | RAGConfig:
    #     """
    #     Run full optimization with optional live progress callback.
        
    #     Args:
    #         n_trials: Total number of optimization iterations
    #         best_one: If True, return single best config; else return all
    #         progress_callback: Optional callback for progress updates
    #         step_size: Iterations per step (for progress updates)
    #         plot_hypervolume: Whether to plot convergence
    #         plot_hypervolume_path: Path to save hypervolume plot
    #         plot_pareto: Whether to plot Pareto front
    #         plot_pareto_path: Path to save Pareto plot
            
    #     Returns:
    #         Best configuration(s) - dict if best_one=False, single config if best_one=True
    #     """
    #     logger.info(f"Starting optimization: {n_trials} trials with step size {step_size}")
        
    #     remaining = n_trials
        
    #     while remaining > 0:
    #         step = min(step_size, remaining)
    #         report = self.run_step(n_iterations=step)
            
    #         if progress_callback is not None:
    #             try:
    #                 progress_callback(report)
    #             except Exception as e:
    #                 logger.warning(f"Progress callback failed: {e}")
                
    #         remaining -= step

    #     # Final report
    #     final_report = self.get_current_progress()
    #     final_report["completed"] = True
    #     final_report["message"] = "Optimization completed successfully!"
        
    #     if progress_callback:
    #         try:
    #             progress_callback(final_report)
    #         except Exception as e:
    #             logger.warning(f"Final progress callback failed: {e}")

    #     # Extract best configurations
    #     best_configs = {
    #         acq: self.search_space.tensor_to_config(
    #             self._last_result.train_x[acq][
    #                 self._last_result.train_obj_true[acq].mean(dim=1).argmax()
    #             ]
    #         )
    #         for acq in self._last_result.train_x.keys()
    #     }

    #     # Generate plots if requested
    #     if plot_hypervolume:
    #         try:
    #             problem = (self.optimization_problem.problem or 
    #                       self.optimization_problem.create_fastmobo_problem())
    #             self._last_result.plot_convergence(problem=problem, save_path=plot_hypervolume_path)
    #             logger.info(f"Hypervolume plot saved to {plot_hypervolume_path or 'display'}")
    #         except Exception as e:
    #             logger.error(f"Failed to plot hypervolume: {e}")

    #     if plot_pareto:
    #         try:
    #             self._last_result.plot_objectives(save_path=plot_pareto_path)
    #             logger.info(f"Pareto plot saved to {plot_pareto_path or 'display'}")
    #         except Exception as e:
    #             logger.error(f"Failed to plot Pareto front: {e}")

    #     logger.success(f"Optimization complete! Evaluated {final_report['total_evaluations']} configurations")
        
    #     # Return single best or all best configs
    #     if best_one and best_configs:
    #         return list(best_configs.values())[0]
    #     return best_configs


    def optimize_with_details(
        self,
        n_trials: int = 50,
        *,
        step_size: int = 5,
        progress_callback: Optional[Callable[[OptimizationProgress], None]] = None,
        plot_hypervolume: bool = False,
        plot_hypervolume_path: Optional[str] = None,
        plot_pareto: bool = False,
        plot_pareto_path: Optional[str] = None,
    ) -> dict[str, RAGConfig]:
        """
        Run optimization with detailed progress tracking and return all configs.
        
        This method runs optimization step-by-step to enable real-time progress updates
        and comprehensive result tracking.
        
        Args:
            n_trials: Total number of optimization iterations
            step_size: Iterations per step (for progress updates)
            progress_callback: Optional callback for progress updates
            plot_hypervolume: Whether to plot convergence
            plot_hypervolume_path: Path to save hypervolume plot
            plot_pareto: Whether to plot Pareto front
            plot_pareto_path: Path to save Pareto plot
            
        Returns:
            Best configuration per acquisition function
        """
        logger.info(f"Starting optimization: {n_trials} trials with step size {step_size}")
        
        remaining = n_trials
        total_completed = 0
        
        while remaining > 0:
            step = min(step_size, remaining)
            
            # Run optimization step
            report = self.run_step(n_iterations=step)
            total_completed += step
            
            # Update report with correct total
            report["iteration"] = total_completed
            report["total_iterations_completed"] = total_completed
            
            # Call progress callback
            if progress_callback is not None:
                try:
                    progress_callback(report)
                except Exception as e:
                    logger.warning(f"Progress callback failed: {e}")
            
            remaining -= step
        
        # Final report
        final_report = self.get_current_progress()
        final_report["completed"] = True
        final_report["iteration"] = total_completed
        final_report["total_iterations_completed"] = total_completed
        final_report["message"] = "Optimization completed successfully!"
        
        if progress_callback:
            try:
                progress_callback(final_report)
            except Exception as e:
                logger.warning(f"Final progress callback failed: {e}")
        
        best_configs = self._extract_best_configs(self._last_result)
        if plot_hypervolume and self._last_result:
            try:
                problem = (self.optimization_problem.problem or 
                        self.optimization_problem.create_fastmobo_problem(evice=device))
                self._last_result.plot_convergence(problem=problem, save_path=plot_hypervolume_path)
                logger.info(f"Hypervolume plot saved to {plot_hypervolume_path or 'display'}")
            except Exception as e:
                logger.error(f"Failed to plot hypervolume: {e}")
        
        if plot_pareto and self._last_result:
            try:
                objective_names = list(self.optimization_problem.objectives)
                self._last_result.plot_objectives(
                    save_path=plot_pareto_path,
                    objective_names=objective_names
                )
                logger.info(f"Pareto plot saved to {plot_pareto_path or 'display'}")
            except Exception as e:
                logger.error(f"Failed to plot Pareto front: {e}")
        
        logger.success(f"Optimization complete! Evaluated {final_report['total_evaluations']} configurations")
        return best_configs


    def optimize(self, 
                n_trials: int = 50, 
                best_one: bool = False,
                *, 
                plot_hypervolume: bool = False,
                plot_hypervolume_path: Optional[str] = None,
                plot_radar: bool = False,
                plot_pareto: bool = False,
                plot_pareto_path: Optional[str] = None,
                objective_names: Optional[list[str]] = None,
                save_result: Optional[bool] = None,
                **kwargs) -> dict[str, RAGConfig] | RAGConfig:
        """
        Run Bayesian optimization to find best RAG configuration
        
        Args:
            n_trials: Number of optimization trials
            best_one: Return single best config if True
            plot_hypervolume: Whether to plot convergence
            plot_hypervolume_path: Path to save hypervolume plot
            plot_pareto: Whether to plot Pareto front
            plot_pareto_path: Path to save Pareto plot
            
        Returns:
            Best configuration per acquisition function
        """
        logger.debug(f"Running {n_trials} optimization trials...")
        
        result: OptimizationResult = self.mobo_optimizer.optimize(
            n_iterations=n_trials, 
            verbose=False,
        )
        logger.warning(f"Optimization complete. Hypervolumes: {result.hypervolumes}")
        
        # Extract best configs for each acquisition function
        best_configs = {}
        
        for acq_func in result.train_x.keys():
            X = result.train_x[acq_func]  
            Y = result.train_obj_true[acq_func] 
            
            avg_performance = Y.mean(dim=1)
            best_idx = avg_performance.argmax().item()
            
            best_config_tensor = X[best_idx]
            
            try:
                best_configs[acq_func] = self.search_space.tensor_to_config(best_config_tensor)
                logger.debug(f"  Successfully decoded Best RAG config for {acq_func} acquisition function")
            except Exception as e:
                logger.error(f"  Failed to decode config for {acq_func}: {e}")
                logger.error(f"  Tensor shape: {best_config_tensor.shape}")
                logger.error(f"  Tensor values: {best_config_tensor}")
        
        if best_one:
            best_configs = list(best_configs.values())[0] if best_configs else None

        logger.success(f"Optimization complete. Hypervolumes: {result.hypervolumes}")

        # Get objective names from evaluator
        objective_names = objective_names or list(self.optimization_problem.objectives) 
        
        if plot_hypervolume:
            problem = self.optimization_problem.problem or self.optimization_problem.create_fastmobo_problem(device=device)
            result.plot_convergence(problem=problem, save_path=plot_hypervolume_path)
        
        if plot_pareto:
            result.plot_objectives(
                save_path=plot_pareto_path, 
                objective_names=objective_names
            )
        
        if plot_radar:
            self.plot_radar_objectives(result, objective_names=objective_names)
        
        if not best_configs:
            logger.warning("optimization failed. No best configs found.")

        # save
        if save_result:
            self.save_result(result, "rag_opt_result.pkl.gz")
        return best_configs


    def plot_radar_objectives(
    self, 
        result: OptimizationResult,
        objective_names: Optional[list[str]] = None,
        save_path: Optional[str] = None
    ):
        """Radar plot of multi-objective performance (mean across samples),
        with automatic label wrapping + spacing to prevent text overlap.
        """

        if objective_names is None:
            if self.optimization_problem:
                objective_names = list(self.optimization_problem.objectives)
            else:
                first_obj = next(iter(result.train_obj_true.values()))
                objective_names = [f"Objective {i+1}" for i in range(first_obj.shape[1])]

        n_objectives = len(objective_names)
        n_methods = len(result.train_obj_true)

        wrapped_names = ["\n".join(textwrap.wrap(name, width=12)) for name in objective_names]

        angles = np.linspace(0, 2 * np.pi, n_objectives, endpoint=False).tolist()
        angles += angles[:1]  # close loop

        fig, axes = plt.subplots(
            1, n_methods,
            figsize=(7 * n_methods, 7),
            subplot_kw=dict(polar=True)
        )

        if n_methods == 1:
            axes = [axes]

        for ax, (method, train_obj) in zip(axes, result.train_obj_true.items()):
            data = train_obj.cpu().numpy()

            # ensure correct shape
            if data.shape[1] != n_objectives:
                logger.warning(
                    f"[Radar Plot] {method}: Expected {n_objectives} objectives, "
                    f"but got shape {data.shape}. Truncating extra metrics."
                )
                data = data[:, :n_objectives]

            min_vals = data.min(axis=0)
            max_vals = data.max(axis=0)
            norm_data = (data - min_vals) / (max_vals - min_vals + 1e-8)

            mean_vals = norm_data.mean(axis=0).tolist()
            mean_vals += mean_vals[:1]

            ax.plot(angles, mean_vals, linewidth=2)
            ax.fill(angles, mean_vals, alpha=0.25)

            ax.set_xticks(angles[:-1])
            ax.set_xticklabels(
                wrapped_names,
                fontsize=11,
                linespacing=1.0
            )

            for label, angle in zip(ax.get_xticklabels(), angles):
                label.set_horizontalalignment(
                    "right" if angle > np.pi/2 and angle < 3*np.pi/2 else "left"
                )

            ax.set_rlabel_position(30)
            ax.set_title(method, fontsize=15, fontweight="bold", pad=20)

        # plt.suptitle(
        #     "Multi-Objective Radar Chart (Mean Normalized Performance)",
        #     fontsize=18,
        #     fontweight="bold",
        #     y=1.05
        # )

        plt.tight_layout()

        if save_path:
            plt.savefig(save_path, dpi=300, bbox_inches="tight")

        plt.show()

    def save_result(self, result: OptimizationResult, path: str):
        """
        Save optimization result (train_x, train_obj_true, hypervolumes, metadata).
        The result can later be reloaded and used to generate all plots.
        """

        payload = {
            "train_x": {k: v.cpu() for k, v in result.train_x.items()},
            "train_obj_true": {k: v.cpu() for k, v in result.train_obj_true.items()},
            "hypervolumes": result.hypervolumes,
            "meta": {
                "timestamp": datetime.datetime.utcnow().isoformat(),
                "objective_names": list(self.optimization_problem.objectives)
                    if self.optimization_problem else None,
                "acquisition_functions": list(result.train_x.keys()),
                "search_space": getattr(self.search_space, "bounds", None),
                "version": "1.0",
            }
        }

        with gzip.open(path, "wb") as f:
            pickle.dump(payload, f)

        logger.success(f"Saved optimization result to {path}")

    def load_result(self, path: str) -> SavedOptimizationResult:
        """
        Load a saved optimization result and rebuild a lightweight result object 
        that can be used for plotting and analysis.
        """

        with gzip.open(path, "rb") as f:
            payload = pickle.load(f)

        result = SavedOptimizationResult(
            train_x=payload["train_x"],
            train_obj_true=payload["train_obj_true"],
            hypervolumes=payload["hypervolumes"],
            meta=payload["meta"],
        )

        logger.success(f"Loaded optimization result from {path}")
        return result



from rag_opt.eval.metrics import MetricResult, MetricCategory, BaseMetric
from concurrent.futures import Future, Executor
from rag_opt.dataset import EvaluationDataset
from typing import Optional, Literal
from rag_opt.llm import RAGLLM, RAGEmbedding
import rag_opt._utils as _utils
from loguru import logger
import torch

NormalizationStrategy = Literal["sum", "softmax", "min-max", "z-score"]

# Default weights for scalarization (NOT used for MOBO)
DEFAULT_WEIGHTS = {
    "cost": 0.3,
    "latency": 0.2,
    "safety": 0.5,
    "alignment": 0.5,
    "response_relevancy": 0.5,
    "context_precision": 0.5,
    "context_recall": 0.3,
    "mrr": 0.3,
    "ndcg": 0.25
}


class RAGEvaluator:
    """
    Multi-objective evaluator for RAG systems.
    
    For Bayesian Optimization:
    - Use evaluate() with normalize=False, return_tensor=True
    - Returns raw objective values (with negation for minimize metrics)
    - Use ref_point property for hypervolume calculation
    """
    
    def __init__(
        self,
        evaluator_llm: Optional[RAGLLM] = None,
        evaluator_embedding: Optional[RAGEmbedding] = None,
        metrics: Optional[list[BaseMetric]] = None,
        *,
        objective_weights: Optional[dict[str, float]] = None,
        auto_initialize_metrics: bool = True,
        executor: Optional[Executor] = None,
        **kwargs
    ):
        self.evaluator_llm = evaluator_llm
        self._metrics: dict[str, BaseMetric] = {}
        self.objective_weights: dict[str, float] = {}
        
        # Load metrics
        if not metrics and auto_initialize_metrics:
            self._initialize_default_metrics(evaluator_llm, evaluator_embedding, **kwargs)
        
        if metrics:
            self.add_metrics(metrics)

        if not self._metrics:
            raise ValueError("No metrics loaded")
        
        self._initialize_weights(objective_weights or DEFAULT_WEIGHTS)
        self._thread_executor = executor or _utils.get_shared_executor()

    @property
    def ref_point(self) -> torch.Tensor:
        """
        Reference point for multi-objective optimization.
        Slightly worse than worst achievable value in transformed space.
        """
        ref_values = []
        for metric in self._metrics.values():
            if metric.negate:
                worst = metric.worst_value if metric.worst_value > 0 else 1.0
                ref_values.append(-worst - 0.1)
            else:
                worst = metric.worst_value if metric.worst_value is not None else 0.0
                ref_values.append(worst - 0.1)
        return torch.tensor(ref_values, dtype=torch.float64)
    
    @property
    def metric_names(self) -> set[str]:
        return set(self._metrics.keys())
    
    @property
    def retrieval_metrics(self) -> dict[str, BaseMetric]:
        return {
            name: metric for name, metric in self._metrics.items()
            if metric.category == MetricCategory.RETRIEVAL
        }
    
    @property
    def generation_metrics(self) -> dict[str, BaseMetric]:
        return {
            name: metric for name, metric in self._metrics.items()
            if metric.category == MetricCategory.GENERATION
        }
    
    @property
    def full_metrics(self) -> dict[str, BaseMetric]:
        return {
            name: metric for name, metric in self._metrics.items()
            if metric.category == MetricCategory.FULL
        }
    
    def _initialize_default_metrics(
        self, 
        llm: RAGLLM, 
        evaluator_embedding: Optional[RAGEmbedding] = None, 
        **kwargs
    ) -> None:
        """Load all default metrics"""
        from rag_opt.eval import all_metrics_factory
        self.add_metrics(all_metrics_factory(llm, evaluator_embedding, **kwargs))
    
    def _initialize_weights(self, weights: dict[str, float]) -> None:
        """Initialize objective weights (only used for scalarization)"""
        if not self._metrics:
            raise ValueError("Cannot initialize weights without metrics")
        
        for name, weight in weights.items():
            if name in self.metric_names:
                self.objective_weights[name] = weight
        
        # Ensure all metrics have weights
        for name in self.metric_names:
            if name not in self.objective_weights:
                logger.warning(f"Metric '{name}' has no weight, defaulting to 0.0")
                self.objective_weights[name] = 0.0
    
    def add_metrics(self, metrics: list[BaseMetric]) -> None:
        """Add multiple metrics"""
        for metric in metrics:
            self.add_metric(metric)
    
    def add_metric(self, metric: BaseMetric, weight: float = 0.0) -> None:
        """Add a single metric with optional weight"""
        if metric.name in self.metric_names:
            logger.warning(f"Overwriting existing metric '{metric.name}'")
        
        self._metrics[metric.name] = metric
        self.objective_weights[metric.name] = weight
    
    def remove_metric(self, name: str) -> None:
        """Remove a metric by name"""
        if name in self._metrics:
            del self._metrics[name]
            self.objective_weights.pop(name, None)
        else:
            logger.warning(f"Cannot remove unknown metric '{name}'")
    
    def evaluate(
        self,
        eval_dataset: EvaluationDataset,
        *,
        return_tensor: bool = True,
        metrics: Optional[dict[str, BaseMetric]] = None,
        normalize: bool = False,
        normalization_strategy: NormalizationStrategy = "sum",
        **kwargs
    ) -> dict[str, MetricResult] | torch.Tensor:
        """
        Evaluate metrics on dataset.
        
        For MOBO: use normalize=False, return_tensor=True
        Returns raw objectives with negation applied.
        
        Args:
            eval_dataset: Dataset to evaluate
            metrics: Specific metrics (defaults to all)
            normalize: Apply normalization (NOT recommended for MOBO)
            return_tensor: Return as tensor
            normalization_strategy: Normalization method
            
        Returns:
            Dictionary of results or tensor of objectives
        """
        metrics_to_eval = metrics or self._metrics
        results: dict[str, MetricResult] = {}
        
        for name, metric in metrics_to_eval.items():
            try:
                result = metric.evaluate(dataset=eval_dataset, **kwargs)
                results[name] = result
            except Exception as e:
                logger.error(f"Error evaluating metric '{name}': {e}")
                results[name] = MetricResult(
                    name=name,
                    value=metric.worst_value,
                    category=metric.category,
                    error=str(e)
                )
        
        if not return_tensor:
            return results
            
        if normalize:
            logger.warning(
                "Normalization enabled. NOT recommended for MOBO as it distorts objective space."
            )
            return self._get_normalized_weighted_scores(
                results, 
                normalization_strategy,
                return_tensor=return_tensor
            )
        
        return self._get_raw_objectives(results, return_tensor=return_tensor)

    def evaluate_batch(
        self,
        eval_datasets: list[EvaluationDataset],
        return_tensor: bool = True,
        **kwargs
    ) -> list[dict[str, MetricResult]] | torch.Tensor:
        """
        Evaluate multiple datasets in parallel.
        
        Args:
            eval_datasets: Datasets to evaluate
            return_tensor: Return stacked tensor
            
        Returns:
            List of results or stacked tensor
        """
        if not eval_datasets:
            return torch.empty(0) if return_tensor else []

        futures: dict[int, Future] = {}
        for index, dataset in enumerate(eval_datasets):
            futures[index] = self._thread_executor.submit(
                self.evaluate, 
                dataset, 
                return_tensor=return_tensor,
                **kwargs
            )

        results: dict[int, dict[str, MetricResult] | torch.Tensor] = {
            index: future.result() for index, future in futures.items()
        }

        if return_tensor:
            return torch.stack([results[i] for i in range(len(eval_datasets))])
        return [results[i] for i in range(len(eval_datasets))]
    
    def _get_raw_objectives(
        self,
        results: dict[str, MetricResult],
        *,
        return_tensor: bool = True
    ) -> torch.Tensor | list[float]:
        """
        Get raw objective vector for MOBO.
        Returns values with negation applied (all "maximize").
        """
        values = []
        
        for name, result in results.items():
            metric = self._metrics[name]
            value = -result.value if metric.negate else result.value
            values.append(value)
        
        return torch.tensor(values, dtype=torch.float64) if return_tensor else values
    
    def _normalize_scores(
        self, 
        scores: list[float], 
        strategy: NormalizationStrategy = "sum"
    ) -> list[float]:
        """Normalize scores using specified strategy"""
        if not scores:
            return []
        
        if strategy == "sum":
            total = sum(scores)
            if total == 0:
                return [1.0 / len(scores)] * len(scores)
            return [w / total for w in scores]
        
        elif strategy == "softmax":
            import math
            max_w = max(scores)
            exp_scores = [math.exp(w - max_w) for w in scores]
            total = sum(exp_scores)
            return [ew / total for ew in exp_scores]
        
        elif strategy == "min-max":
            min_val, max_val = min(scores), max(scores)
            if max_val == min_val:
                return [1.0 / len(scores)] * len(scores)
            scaled = [(w - min_val) / (max_val - min_val) for w in scores]
            total = sum(scaled)
            return [s / total for s in scaled] if total > 0 else scaled
        
        elif strategy == "z-score":
            import statistics
            if len(scores) < 2:
                return scores
            mean = statistics.mean(scores)
            std = statistics.stdev(scores)
            if std == 0:
                return [1.0 / len(scores)] * len(scores)
            z_scores = [(w - mean) / std for w in scores]
            min_z = min(z_scores)
            shifted = [z - min_z for z in z_scores]
            total = sum(shifted)
            return [s / total for s in shifted] if total > 0 else [1.0 / len(scores)] * len(scores)
        
        raise ValueError(f"Unknown normalization strategy: {strategy}")
    
    def _get_normalized_weighted_scores(
        self, 
        results: dict[str, MetricResult],
        normalization_strategy: str = "sum",
        apply_weights: bool = True,
        *,
        return_tensor: bool = True
    ) -> torch.Tensor | list[float]:
        """
        Get weighted normalized scores for scalarization.
        WARNING: NOT for MOBO - only for single-objective or visualization.
        """
        values = []
        weights = []

        for name, result in results.items():
            metric = self._metrics[name]
            value = -result.value if metric.negate else result.value
            values.append(value)
            weights.append(self.objective_weights.get(name, 0.0))

        normalized_values = self._normalize_scores(values, strategy=normalization_strategy)

        if apply_weights:
            normalized_weights = self._normalize_scores(weights, strategy="sum")
            weighted_values = [v * w for v, w in zip(normalized_values, normalized_weights)]
        else:
            weighted_values = normalized_values

        return torch.tensor(weighted_values, dtype=torch.float64) if return_tensor else weighted_values
    
    def compute_objective_score(
        self,
        results: dict[str, MetricResult],
        normalization_strategy: NormalizationStrategy = "sum"
    ) -> float:
        """
        Compute single aggregated score for scalarization.
        NOTE: NOT used for MOBO - only for debugging/visualization.
        """
        values = []
        weights = []
        
        for name, result in results.items():
            metric = self._metrics[name]
            value = -result.value if metric.negate else result.value
            values.append(value)
            weights.append(self.objective_weights.get(name, 0.0))
        
        normalized_weights = self._normalize_scores(weights, strategy=normalization_strategy)
        return sum(v * w for v, w in zip(values, normalized_weights))
    
    def evaluate_retrieval(
        self,
        eval_dataset: EvaluationDataset,
        **kwargs
    ) -> dict[str, MetricResult]:
        """Evaluate only retrieval metrics"""
        return self.evaluate(
            eval_dataset, 
            metrics=self.retrieval_metrics,
            return_tensor=False,
            **kwargs
        )
    
    def evaluate_generation(
        self,
        eval_dataset: EvaluationDataset,
        **kwargs
    ) -> dict[str, MetricResult]:
        """Evaluate only generation metrics"""
        return self.evaluate(
            eval_dataset,
            metrics=self.generation_metrics,
            return_tensor=False,
            **kwargs
        )
    
    def evaluate_full(
        self,
        eval_dataset: EvaluationDataset,
        **kwargs
    ) -> dict[str, MetricResult]:
        """Evaluate full pipeline metrics"""
        return self.evaluate(
            eval_dataset,
            metrics=self.full_metrics,
            return_tensor=False,
            **kwargs
        )
    
    def available_metrics(self) -> list[str]:
        """List all available metric names"""
        return list(self.metric_names)


from botorch.utils.sampling import draw_sobol_samples
from typing import Optional
from rag_opt._config import (
    RAGConfig,
    EmbeddingConfig,
    EmbeddingModel,
    LLMConfig,
    LLMModel,
    RerankerConfig,
    RerankerModel,
    VectorStoreConfig,
    VectorStoreItem,
)
from abc import ABC, abstractmethod
from enum import Enum, auto
from loguru import logger
import numpy as np
import random
import torch


class SamplerType(Enum):
    """Supported sampling methods"""
    SOBOL = auto()
    RANDOM = auto()


class SamplingMixin(ABC):
    """Mixin for sampling RAG configurations with different strategies"""

    @abstractmethod
    def _get_hyperparameters(self) -> dict:
        """Get hyperparameter configuration"""
        raise NotImplementedError

    def _get_expanded_hyperparameters(self) -> dict:
        """Expand hyperparameters to include model selection dimensions"""
        expanded = {}
        base_params = self._get_hyperparameters()

        for param_name, config in base_params.items():
            if not config:
                continue
                
            expanded[param_name] = config.copy()
            choices = config.get("choices", {})
            
            if not isinstance(choices, dict):
                continue
            
            # Add model index parameter for model-based configs
            for choice_key, choice in choices.items():
                if self._is_model_based_config(choice):
                    models = self._get_models_from_config(choice)
                    if models:
                        model_param_name = f"{param_name}_model_idx"
                        expanded[model_param_name] = {
                            "searchspace_type": "categorical",
                            "choices": {str(i): i for i in range(len(models))},
                            "bounds": [0, len(models) - 1],
                            "dtype": int
                        }
                        break

        return expanded

    def _get_models_from_config(self, config) -> list[str]:
        """Extract models list from config (handles both Config objects and lists)"""
        if isinstance(config, list):
            return config
        elif hasattr(config, 'models'):
            return config.models
        return []

    def _is_model_based_config(self, config) -> bool:
        """Check if config has models (either as attribute or is a list)"""
        return isinstance(config, list) or (
            isinstance(config, (EmbeddingConfig, LLMConfig, RerankerConfig)) and 
            hasattr(config, 'models')
        )

    def get_parameter_bounds(self) -> torch.Tensor:
        """Get normalized [0,1] bounds for BoTorch"""
        expanded_params = self._get_expanded_hyperparameters()
        n_params = len(expanded_params)
        
        if n_params == 0:
            return torch.empty((2, 0), dtype=torch.float64)
        
        # All parameters normalized to [0, 1] for BoTorch
        bounds = torch.zeros((2, n_params), dtype=torch.float64)
        bounds[1, :] = 1.0
        
        return bounds

    def _normalize_value(self, value: float, lower: float, upper: float) -> float:
        """Normalize value from [lower, upper] to [0, 1]"""
        if upper == lower:
            return 0.5
        return (value - lower) / (upper - lower)

    def _denormalize_value(self, normalized: float, lower: float, upper: float, dtype: type) -> float:
        """Denormalize value from [0, 1] to [lower, upper]"""
        normalized = max(0.0, min(1.0, normalized))
        value = lower + normalized * (upper - lower)
        return value if dtype == float else int(round(value))

    def config_to_tensor(self, config: RAGConfig) -> torch.Tensor:
        """Encode RAGConfig to normalized tensor [0,1]"""
        expanded_params = self._get_expanded_hyperparameters()
        base_params = self._get_hyperparameters()

        sample = torch.zeros(len(expanded_params), dtype=torch.float64)
        param_idx = 0

        for param_name, param_config in base_params.items():
            config_value = getattr(config, param_name)
            param_type = param_config.get("searchspace_type")
            choices = param_config.get("choices")
            dtype = param_config.get("dtype")
            bounds = param_config.get("bounds", [0, 1])

            if param_type == 'continuous':
                raw_value = float(config_value) if dtype == float else int(config_value)
                sample[param_idx] = self._normalize_value(raw_value, bounds[0], bounds[1])

            elif param_type == 'categorical':
                if isinstance(choices, dict):
                    choices_list = list(choices.values())
                    choice_idx = self._find_matching_choice(config_value, choices_list)
                    n_choices = len(choices_list)
                    sample[param_idx] = choice_idx / (n_choices - 1) if n_choices > 1 else 0.5

                    # Handle model index
                    model_param_name = f"{param_name}_model_idx"
                    if model_param_name in expanded_params:
                        param_idx += 1
                        model_idx = self._extract_model_index(config_value, choices_list[choice_idx])
                        models = self._get_models_from_config(choices_list[choice_idx])
                        n_models = len(models)
                        sample[param_idx] = model_idx / (n_models - 1) if n_models > 1 else 0.5
                else:
                    sample[param_idx] = 0.5

            elif param_type == 'boolean':
                sample[param_idx] = 1.0 if config_value else 0.0

            param_idx += 1

        return sample
    
    def tensor_to_config(self, tensor: torch.Tensor) -> RAGConfig:
        """Decode tensor to RAGConfig"""
        return self.decode_sample_to_rag_config(tensor)

    def configs_to_tensor(self, configs: list[RAGConfig]) -> torch.Tensor:
        """Convert list of configs to tensor"""
        if not configs:
            expanded_params = self._get_expanded_hyperparameters()
            return torch.empty((0, len(expanded_params)), dtype=torch.float64)

        return torch.stack([self.config_to_tensor(config) for config in configs])

    def _extract_model_index(self, config_value, choice_config) -> int:
        """Extract model index from config"""
        # Handle gateway list case
        if isinstance(choice_config, list):
            if hasattr(config_value, 'model'):
                try:
                    return choice_config.index(config_value.model)
                except (ValueError, AttributeError):
                    return 0
            return 0
        
        # Handle standard config case
        if not (hasattr(config_value, 'model') and hasattr(choice_config, 'models')):
            return 0
        try:
            return choice_config.models.index(config_value.model)
        except (ValueError, AttributeError):
            return 0

    def _find_matching_choice(self, config_value, choices_list: list) -> int:
        """Find matching choice index"""
        for idx, choice in enumerate(choices_list):
            # Match gateway list
            if isinstance(choice, list):
                if hasattr(config_value, 'model') and config_value.model in choice:
                    return idx
            # Match by provider and model
            elif hasattr(config_value, 'provider') and hasattr(config_value, 'model'):
                if (config_value.provider == choice.provider and
                    hasattr(choice, 'models') and
                    config_value.model in choice.models):
                    return idx
            # Match by provider only
            elif hasattr(config_value, 'provider') and hasattr(choice, 'provider'):
                if config_value.provider == choice.provider:
                    return idx
            # Direct match
            elif config_value == choice:
                return idx
        return 0

    def decode_sample_to_rag_config(self, sample: torch.Tensor) -> RAGConfig:
        """Decode normalized sample to RAGConfig"""
        decoded = {}
        base_params = self._get_hyperparameters()
        expanded_params = self._get_expanded_hyperparameters()

        param_idx = 0
        for param_name, config in base_params.items():
            value = max(0.0, min(1.0, sample[param_idx].item()))
            param_type = config.get("searchspace_type")
            choices = config.get("choices")
            dtype = config.get("dtype")
            bounds = config.get("bounds", [0, 1])

            if param_type == 'continuous':
                decoded[param_name] = self._denormalize_value(value, bounds[0], bounds[1], dtype)

            elif param_type == 'categorical':
                if isinstance(choices, dict):
                    choices_list = list(choices.values())
                    if not choices_list:
                        raise ValueError(f"No choices for parameter '{param_name}'")
                    
                    choice_idx = min(int(round(value * (len(choices_list) - 1))), len(choices_list) - 1)
                    choice = choices_list[choice_idx]

                    # Get model index if present
                    model_param_name = f"{param_name}_model_idx"
                    model_idx = 0
                    if model_param_name in expanded_params:
                        param_idx += 1
                        model_idx_value = max(0.0, min(1.0, sample[param_idx].item()))
                        models = self._get_models_from_config(choice)
                        model_idx = int(round(model_idx_value * (len(models) - 1)))

                    decoded[param_name] = self._select_model_from_config(choice, model_idx, list(choices.keys())[choice_idx], param_name)

                elif isinstance(choices, list):
                    if not choices:
                        raise ValueError(f"No choices for parameter '{param_name}'")
                    choice_idx = min(int(round(value * (len(choices) - 1))), len(choices) - 1)
                    decoded[param_name] = choices[choice_idx]
                else:
                    decoded[param_name] = choices

            elif param_type == 'boolean':
                decoded[param_name] = value >= 0.5

            param_idx += 1

        return RAGConfig(**decoded)

    def _select_model_from_config(self, config, model_idx: int, provider_key: str = None, param_name: str = None):
        """Select specific model from config (handles both Config objects and gateway lists)"""
        # Handle gateway list case
        if isinstance(config, list):
            if not config:
                raise ValueError("Empty model list in gateway configuration")
            model_idx = max(0, min(model_idx, len(config) - 1))
            selected_model = config[model_idx]
            
            # Parse provider/model from gateway format (e.g., "openai/gpt-4o")
            if '/' in selected_model:
                provider, model = selected_model.split('/', 1)
            else:
                provider = provider_key or "gateway"
                model = selected_model
            
            # Return appropriate model type based on parameter name
            if param_name == 'embedding':
                return EmbeddingModel(
                    provider=provider,
                    model=model,
                    api_key=None,
                    api_base=None,
                    pricing=None
                )
            elif param_name == 'llm':
                return LLMModel(
                    provider=provider,
                    model=model,
                    api_key=None,
                    api_base=None,
                    pricing=None
                )
            elif param_name == 'reranker':
                return RerankerModel(
                    provider=provider,
                    model=model,
                    api_key=None,
                    api_base=None,
                    pricing=None
                )
            else:
                # Default to EmbeddingModel if unknown
                logger.warning(f"Unknown param_name '{param_name}' for gateway config, defaulting to EmbeddingModel")
                return EmbeddingModel(
                    provider=provider,
                    model=model,
                    api_key=None,
                    api_base=None,
                    pricing=None
                )
        
        # Handle VectorStoreConfig case
        if isinstance(config, VectorStoreConfig):
            return VectorStoreItem(
                provider=config.provider,
                index_name=config.index_name,
                api_key=config.api_key,
                pricing=config.pricing
            )
        
        # Handle standard Config objects
        models = self._get_models_from_config(config)
        if not models:
            return config

        model_idx = max(0, min(model_idx, len(models) - 1))
        selected_model = models[model_idx]

        if isinstance(config, EmbeddingConfig):
            return EmbeddingModel(
                provider=config.provider,
                model=selected_model,
                api_key=config.api_key,
                api_base=config.api_base,
                pricing=config.pricing.get(selected_model) if config.pricing else None
            )
        elif isinstance(config, LLMConfig):
            return LLMModel(
                provider=config.provider,
                model=selected_model,
                api_key=config.api_key,
                api_base=config.api_base,
                pricing=config.pricing.get(selected_model) if config.pricing else None
            )
        elif isinstance(config, RerankerConfig):
            return RerankerModel(
                provider=config.provider,
                model=selected_model,
                api_key=config.api_key,
                api_base=config.api_base,
                pricing=config.pricing.get(selected_model) if config.pricing else None
            )
        else:
            logger.warning(f"Unknown config type: {type(config)}")
            return config

    def sample(
        self,
        n_samples: int = 1,
        sampler_type: SamplerType = SamplerType.SOBOL,
        seed: Optional[int] = None
    ) -> list[RAGConfig]:
        """Sample RAG configurations"""
        if seed is not None:
            torch.manual_seed(seed)
            np.random.seed(seed)
            random.seed(seed)

        bounds = self.get_parameter_bounds()
        n_params = bounds.shape[1]

        if n_params == 0:
            return []

        if sampler_type == SamplerType.SOBOL:
            samples = draw_sobol_samples(
                bounds=bounds,
                n=n_samples,
                q=1,
                seed=seed
            ).squeeze(1)
        elif sampler_type == SamplerType.RANDOM:
            samples = torch.rand(n_samples, n_params, dtype=torch.float64)
        else:
            raise ValueError(f"Unknown sampler type: {sampler_type}")

        return [self.decode_sample_to_rag_config(samples[i]) for i in range(n_samples)]

    def sample_batch(
        self,
        batch_size: int,
        sampler_type: SamplerType = SamplerType.SOBOL,
        seed: Optional[int] = None
    ) -> list[RAGConfig]:
        """Convenience method for batch sampling"""
        return self.sample(n_samples=batch_size, sampler_type=sampler_type, seed=seed)


from langchain_core.documents import BaseDocumentCompressor, Document
from abc import ABC, abstractmethod
from loguru import logger
from typing import Optional


class BaseReranker(ABC):
    """Abstract base class for all rerankers"""
    
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.model = self._initialize_model()
    
    @abstractmethod
    def _initialize_model(self) -> BaseDocumentCompressor:
        """Initialize the specific reranker model"""
        pass
    
    @abstractmethod
    def rerank(self, query: str, documents: list[Document], top_k: int = 10) -> list[Document]:
        """Rerank documents based on query relevance"""
        pass


class HuggingFaceCrossEncoderReranker(BaseReranker):
    """Cross-encoder reranker using HuggingFace models"""
    
    def _initialize_model(self):
        from langchain_community.cross_encoders import HuggingFaceCrossEncoder
        model_name = self.kwargs.get("model_name", "BAAI/bge-reranker-base")
        return HuggingFaceCrossEncoder(model_name=model_name) 
    
    def rerank(self, query: str, documents: list[Document], top_k: int = 10) -> list[Document]:
        pairs = [(query, doc.page_content) for doc in documents]
        scores = self.model.score(pairs)
        
        scored_docs = sorted(zip(documents, scores), key=lambda x: x[1], reverse=True)
        return [doc for doc, _ in scored_docs[:top_k]]


class CohereReranker(BaseReranker):
    """Cohere reranker implementation"""
    
    def _initialize_model(self):
        from langchain.retrievers.document_compressors import CohereRerank
        
        api_key = self.kwargs.get("api_key")
        if not api_key:
            raise ValueError("api_key is required for Cohere reranker")
        
        model = self.kwargs.get("model_name", "rerank-english-v3.0")
        return CohereRerank(cohere_api_key=api_key, model=model, top_n=100)
    
    def rerank(self, query: str, documents: list[Document], top_k: int = 10) -> list[Document]:
        return self.model.compress_documents(documents, query)[:top_k]


class FlashRankReranker(BaseReranker):
    """FlashRank reranker implementation"""
    
    def _initialize_model(self):
        from flashrank import Ranker
        model = self.kwargs.get("model_name", "ms-marco-MiniLM-L-12-v2")
        return Ranker(model_name=model)
    
    def rerank(self, query: str, documents: list[Document], top_k: int = 10) -> list[Document]:
        passages = [{"id": i, "text": doc.page_content} for i, doc in enumerate(documents)]
        results = self.model.rerank(query, passages)
        return [documents[r["id"]] for r in results[:top_k]]


class JinaReranker(BaseReranker):
    """Jina AI reranker implementation"""
    
    def _initialize_model(self):
        from langchain_community.document_compressors import JinaRerank
        
        api_key = self.kwargs.get("api_key")
        if not api_key:
            raise ValueError("api_key is required for Jina reranker")
        
        model = self.kwargs.get("model_name", "jina-reranker-v1-base-en")
        return JinaRerank(jina_api_key=api_key, model=model, top_n=100)
    
    def rerank(self, query: str, documents: list[Document], top_k: int = 10) -> list[Document]:
        return self.model.compress_documents(documents, query)[:top_k]


class VoyageAIReranker(BaseReranker):
    """Voyage AI reranker implementation"""
    
    def _initialize_model(self):
        from langchain_voyageai import VoyageAIRerank
        
        api_key = self.kwargs.get("api_key")
        if not api_key:
            raise ValueError("api_key is required for Voyage AI reranker")
        
        model = self.kwargs.get("model_name", "rerank-lite-1")
        return VoyageAIRerank(voyageai_api_key=api_key, model=model, top_k=100)
    
    def rerank(self, query: str, documents: list[Document], top_k: int = 10) -> list[Document]:
        return self.model.compress_documents(documents, query)[:top_k]


class BM25Reranker(BaseReranker):
    """BM25 statistical reranker"""
    
    def _initialize_model(self):
        return None  # Initialized on-the-fly during rerank
    
    def rerank(self, query: str, documents: list[Document], top_k: int = 10) -> list[Document]:
        from rank_bm25 import BM25Okapi
        
        # Tokenize corpus and query
        tokenized_corpus = [doc.page_content.lower().split() for doc in documents]
        tokenized_query = query.lower().split()
        
        # Get BM25 scores
        bm25 = BM25Okapi(tokenized_corpus)
        scores = bm25.get_scores(tokenized_query)
        
        # Return top-k documents
        scored_docs = sorted(zip(documents, scores), key=lambda x: x[1], reverse=True)
        return [doc for doc, _ in scored_docs[:top_k]]


class PineconeReranker(BaseReranker):
    """Pinecone Inference reranker"""
    
    def _initialize_model(self):
        from langchain_pinecone import PineconeRerank
        
        api_key = self.kwargs.get("api_key")
        if not api_key:
            raise ValueError("api_key is required for Pinecone reranker")
        
        model = self.kwargs.get("model_name", "bge-reranker-v2-m3")
        return PineconeRerank(api_key=api_key, model=model)
    
    def rerank(self, query: str, documents: list[Document], top_k: int = 10) -> list[Document]:
        docs_text = [doc.page_content for doc in documents]
        results = self.model.rerank(query=query, documents=docs_text, top_n=top_k)
        return [documents[r.index] for r in results]


# Registry of available rerankers
RERANKER_REGISTRY = {
    "cohere": CohereReranker,
    "flashrank": FlashRankReranker,
    "jina": JinaReranker,
    "voyageai": VoyageAIReranker,
    "bm25": BM25Reranker,
    "pinecone": PineconeReranker,
    "huggingface": HuggingFaceCrossEncoderReranker,
}


def init_reranker(
    provider: str,
    model_name: Optional[str] = None,
    api_key: Optional[str] = None,
    **kwargs
) -> BaseReranker:
    """
    Initialize a reranker based on provider.
    
    Args:
        provider: Reranker provider (cohere, flashrank, jina, voyageai, bm25, pinecone, huggingface)
        model_name: Specific model to use (optional)
        api_key: API key for cloud providers
        **kwargs: Additional provider-specific arguments
    
    Returns:
        Initialized reranker instance
    
    Examples:
        >>> reranker = init_reranker("cohere", api_key="your-key")
        >>> reranker = init_reranker("huggingface", model_name="BAAI/bge-reranker-base")
        >>> reranker = init_reranker("bm25")  # No API key needed
    """
    provider = provider.lower()
    
    if provider not in RERANKER_REGISTRY:
        available = ", ".join(RERANKER_REGISTRY.keys())
        raise ValueError(f"Unsupported reranker: {provider}. Available: {available}")
    
    # Build config
    config = kwargs.copy()
    if model_name:
        config["model_name"] = model_name
    if api_key:
        config["api_key"] = api_key
    
    # Initialize reranker
    reranker_class = RERANKER_REGISTRY[provider]
    reranker = reranker_class(**config)
    
    logger.success(f"{provider.capitalize()} reranker loaded successfully")
    return reranker


from langchain_core.prompts import get_template_variables
from concurrent.futures import ThreadPoolExecutor
from loguru import logger
import functools
import atexit
import re 


@functools.lru_cache(maxsize=1)
def get_shared_executor(max_workers:int=5) -> ThreadPoolExecutor:
    """ Return a lazily created, shared ThreadPoolExecutor.

    This ensures that only one executor instance is created and reused 
    throughout the lifetime of the program. The executor is automatically 
    shut down at program exit 

    Args:
        max_workers (int): Maximum number of worker threads for the executor.
                           Only applied on the first call.

    Returns:
        ThreadPoolExecutor: A shared executor instance.
    """
    executor = ThreadPoolExecutor(max_workers=max_workers)
    atexit.register(executor.shutdown, wait=True)
    return executor


def extract_num_from_text(text: str) -> int | None:
    """
    Fallback function to extract a number from text
    """
    if not text:
        return None
    
    match = re.search(r"-?\d+", text)
    if match:
        try:
            return int(match.group())
        except ValueError:
            logger.error(f"Failed to parse number from text: {text}")
            return None
    return None


def validate_prompt(prompt_template:str,reference_prompt:str, raise_error:bool=True) -> bool:
    prompt_vars = get_template_variables(reference_prompt, "f-string")
    vars = get_template_variables(prompt_template, "f-string")
    needed_vars = set(prompt_vars) - set(vars)
    diff = set(vars) ^ set(prompt_vars)
    if needed_vars:
        logger.error(f"Your prompt is missing the following variables: {needed_vars}")
        if raise_error:
            raise ValueError(f"Your prompt is missing the following variables: {needed_vars}")
        return False
    
    if diff:
        logger.error(f"Your prompt contains extra variables: {diff} \n your prompt must only contain the following variables only:  {prompt_vars}")
        
        if raise_error:
            raise ValueError(f"Your prompt contains extra variables: {diff}")
        return False
    
    return True 


"""
Evaluation metrics for Embedding + VectorStore + Reranker (Optimized)
"""
from rag_opt.eval.metrics.base import BaseMetric, MetricCategory
from rag_opt._prompts import CONTEXT_PRECISION_PROMPT, CONTEXT_RECALL_PROMPT
from langchain_core.messages import BaseMessage
from rag_opt.dataset import EvaluationDataset
import rag_opt._utils as _utils
from abc import ABC
from loguru import logger
from rag_opt.llm import RAGEmbedding
import math
import json

TEXT_SIMILARITY_THRESHOLD = 0.8


class RetrievalMetrics(BaseMetric, ABC):
    """Base class for retrieval metrics with optimized embedding caching"""
    category: MetricCategory = MetricCategory.RETRIEVAL
    is_llm_based: bool = False
    is_embedding_based: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._limit_contexts = kwargs.get("limit_contexts", 3)
        self.embedding_model: RAGEmbedding = kwargs.get("embedding_model", None)
        
        # Token management configuration
        self.max_prompt_tokens = kwargs.get("max_prompt_tokens", 16000)
        self.batch_size = kwargs.get("batch_size", 10)
        self.max_context_length = kwargs.get("max_context_length", 4000)
        
        # Embedding cache to avoid recomputation (CRITICAL OPTIMIZATION)
        self._embedding_cache = {}
        
        # Token encoder cache
        self._token_encoder = None
        try:
            import tiktoken
            self._token_encoder = tiktoken.get_encoding("cl100k_base")
        except Exception as e:
            logger.warning(f"tiktoken not available: {e}. Using approximation.")

        if self.is_embedding_based and not self.embedding_model:
            logger.error(f"Embedding is required to evaluate {self.name}")
            raise ValueError(f"Embedding is required to evaluate {self.name}")
    
    @property
    def limit_contexts(self):
        return self._limit_contexts

    @limit_contexts.setter
    def limit_contexts(self, value: int):
        self._limit_contexts = value

    def _estimate_tokens(self, text: str) -> int:
        """Estimate token count with cached encoder"""
        if self._token_encoder:
            return len(self._token_encoder.encode(text))
        return len(text) // 4

    def _truncate_text(self, text: str, max_chars: int) -> str:
        """Truncate text to max_chars, trying to preserve sentence boundaries"""
        if len(text) <= max_chars:
            return text
        
        truncated = text[:max_chars]
        last_period = truncated.rfind('. ')
        last_newline = truncated.rfind('\n')
        boundary = max(last_period, last_newline)
        
        if boundary > max_chars * 0.8:
            return truncated[:boundary + 1] + "..."
        
        return truncated + "..."

    def _fit_contexts_to_token_limit(
        self, 
        contexts: list[str], 
        template_overhead: str,
        max_tokens: int = None
    ) -> list[str]:
        """Dynamically adjust contexts to fit within token limit"""
        if max_tokens is None:
            max_tokens = self.max_prompt_tokens
        
        template_tokens = self._estimate_tokens(template_overhead)
        available_tokens = max_tokens - template_tokens - 100
        
        if available_tokens <= 0:
            logger.error("Template overhead exceeds token limit!")
            return []
        
        # Try with original limit_contexts
        limited_contexts = contexts[:self.limit_contexts]
        contexts_text = "\n".join(limited_contexts)
        contexts_tokens = self._estimate_tokens(contexts_text)
        
        if contexts_tokens <= available_tokens:
            return limited_contexts
        
        # Reduce number of contexts
        num_contexts = self.limit_contexts
        while num_contexts > 0:
            limited_contexts = contexts[:num_contexts]
            contexts_text = "\n".join(limited_contexts)
            contexts_tokens = self._estimate_tokens(contexts_text)
            
            if contexts_tokens <= available_tokens:
                return limited_contexts
            
            num_contexts -= 1
        
        # Truncate single context if needed
        if contexts:
            max_chars = available_tokens * 4
            truncated = self._truncate_text(contexts[0], max_chars)
            logger.warning(
                f"Single context exceeds limit. Truncated from "
                f"{len(contexts[0])} to {len(truncated)} characters"
            )
            return [truncated]
        
        return []

    def _build_prompt_with_token_limit(
        self, 
        template: str, 
        contexts: list[str],
        **kwargs
    ) -> str:
        """Build prompt ensuring it fits within token limit"""
        template_overhead = template.format(context="[CONTEXTS_PLACEHOLDER]", **kwargs)
        fitted_contexts = self._fit_contexts_to_token_limit(contexts, template_overhead)
        return template.format(context=fitted_contexts, **kwargs)

    def _parse_llm_responses(self, responses: list[BaseMessage]) -> list[float]:
        if self.is_llm_based:
            raise NotImplementedError
    
    def _verify_with_llm(self, prompts: list[str]) -> list[float]:
        """Common LLM verification with batching"""
        if not self.llm:
            logger.error(f"LLM is required to evaluate {self.name}")
            raise ValueError(f"LLM is required to evaluate {self.name}")
        
        processed_prompts = []
        for idx, prompt in enumerate(prompts):
            token_count = self._estimate_tokens(prompt)
            
            if token_count > self.max_prompt_tokens:
                logger.warning(
                    f"Prompt {idx} exceeds token limit: {token_count} > {self.max_prompt_tokens}. Truncating..."
                )
                prompt = self._truncate_prompt_fallback(prompt, self.max_prompt_tokens)
            
            processed_prompts.append(prompt)
        
        all_responses = []
        total_batches = (len(processed_prompts) - 1) // self.batch_size + 1
        
        for batch_idx in range(0, len(processed_prompts), self.batch_size):
            batch = processed_prompts[batch_idx:batch_idx + self.batch_size]
            logger.debug(
                f"Processing batch {batch_idx//self.batch_size + 1}/{total_batches} ({len(batch)} prompts)"
            )
            
            try:
                responses = self.llm.batch(batch)
                all_responses.extend(responses)
            except Exception as e:
                logger.error(f"Batch processing failed: {e}")
                all_responses.extend([None] * len(batch))
        
        return self._parse_llm_responses(all_responses)

    def _truncate_prompt_fallback(self, prompt: str, max_tokens: int) -> str:
        """Fallback truncation for prompts that exceed limit"""
        max_chars = max_tokens * 4
        
        if len(prompt) <= max_chars:
            return prompt
        
        if "Context:" in prompt or "context:" in prompt:
            parts = prompt.split("\n")
            header_parts = []
            context_parts = []
            footer_parts = []
            
            in_context = False
            for part in parts:
                if "context:" in part.lower():
                    in_context = True
                    header_parts.append(part)
                elif "question:" in part.lower() or "answer:" in part.lower():
                    in_context = False
                    footer_parts.append(part)
                elif in_context:
                    context_parts.append(part)
                elif footer_parts:
                    footer_parts.append(part)
                else:
                    header_parts.append(part)
            
            header_text = "\n".join(header_parts)
            footer_text = "\n".join(footer_parts)
            overhead = len(header_text) + len(footer_text) + 100
            available_for_context = max(max_chars - overhead, max_chars // 2)
            
            context_text = "\n".join(context_parts)
            if len(context_text) > available_for_context:
                context_text = self._truncate_text(context_text, available_for_context)
                context_text += "\n[... contexts truncated due to length ...]"
            
            return f"{header_text}\n{context_text}\n{footer_text}"
        
        return self._truncate_text(prompt, max_chars) + "\n[Content truncated to fit token limit]"

    # ========== EMBEDDING OPTIMIZATION METHODS ==========
    
    def _get_cached_embedding(self, text: str) -> list[float]:
        """Get embedding with caching to avoid redundant API calls"""
        text_clean = text.strip()
        
        if text_clean not in self._embedding_cache:
            self._embedding_cache[text_clean] = self.embedding_model.embed_query(text_clean)
        
        return self._embedding_cache[text_clean]
    
    def _precompute_embeddings(self, contexts_list: list[list[str]]):
        """Precompute all embeddings in batch before evaluation (MAJOR SPEEDUP)"""
        if not self.embedding_model:
            return
        
        # Collect all unique contexts
        unique_contexts = set()
        for contexts in contexts_list:
            for ctx in contexts:
                unique_contexts.add(ctx.strip())
        
        unique_contexts = list(unique_contexts)
        
        if not unique_contexts:
            return
        
        logger.info(f"Precomputing {len(unique_contexts)} unique contexts...")
        
        try:
            # Try batch embedding first
            if hasattr(self.embedding_model, 'embed_documents'):
                embeddings = self.embedding_model.embed_documents(unique_contexts)
                for ctx, emb in zip(unique_contexts, embeddings):
                    self._embedding_cache[ctx] = emb
                logger.debug(f"Batch embedding completed for {len(unique_contexts)} contexts")
            else:
                # Fallback: sequential with progress logging
                for i, ctx in enumerate(unique_contexts):
                    if i % 10 == 0:
                        logger.debug(f"Embedding progress: {i}/{len(unique_contexts)}")
                    self._embedding_cache[ctx] = self.embedding_model.embed_query(ctx)
                logger.info(f"Sequential embedding completed for {len(unique_contexts)} contexts")
        except Exception as e:
            logger.warning(f"Batch embedding failed: {e}. Will compute on-demand.")
    
    def _cosine_similarity(self, vec1: list[float], vec2: list[float]) -> float:
        """Calculate cosine similarity between two vectors"""
        dot_product = sum(a * b for a, b in zip(vec1, vec2))
        magnitude1 = math.sqrt(sum(a * a for a in vec1))
        magnitude2 = math.sqrt(sum(b * b for b in vec2))
        
        if magnitude1 == 0 or magnitude2 == 0:
            return 0.0
        
        return dot_product / (magnitude1 * magnitude2)
    
    def _is_match(self, retrieved_ctx: str, ground_truth_contexts: list[str]) -> bool:
        """Check if retrieved context matches any ground truth (uses cached embeddings)"""
        retrieved_clean = retrieved_ctx.strip()

        # Quick exact match checks first (no API calls)
        for gt_ctx in ground_truth_contexts:
            gt_clean = gt_ctx.strip()
            
            if retrieved_clean == gt_clean or retrieved_clean.lower() == gt_clean.lower():
                return True
            
            # Word-level overlap check
            r_words = set(retrieved_clean.lower().split())
            g_words = set(gt_clean.lower().split())
            overlap = len(r_words & g_words)
            
            if overlap > 0:
                word_overlap = overlap / max(len(r_words | g_words), 1)
                if word_overlap >= 0.6:
                    return True
        
        # Semantic similarity check with cached embeddings
        if self.embedding_model:
            try:
                retrieved_embedding = self._get_cached_embedding(retrieved_clean)
                
                for gt_ctx in ground_truth_contexts:
                    gt_clean = gt_ctx.strip()
                    gt_embedding = self._get_cached_embedding(gt_clean)
                    
                    cosine_sim = self._cosine_similarity(retrieved_embedding, gt_embedding)
                    if cosine_sim >= 0.7:
                        return True
            except Exception as e:
                logger.warning(f"Embedding-based similarity failed: {e}")
        
        return False


# *************************
# Context-Based Metrics
# *************************
class ContextPrecision(RetrievalMetrics):
    """Context Precision: Measures relevance of retrieved documents using Average Precision"""
    
    name: str = "context_precision"
    _prompt_template: str = CONTEXT_PRECISION_PROMPT
    is_llm_based: bool = True
    
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("name", "context_precision")
        kwargs.setdefault("category", MetricCategory.RETRIEVAL)
        super().__init__(*args, **kwargs)
    
    def _calculate_context_precision(self, contexts_verifications: list[int]) -> list[float]:
        """Calculate context precision using Average Precision (AP) formula"""
        if not contexts_verifications:
            logger.warning("No relevant contexts found")
            return []

        tot_sum = sum(contexts_verifications)
        
        if tot_sum == 0:
            num = 0.0
        else:
            num = sum([
                sum(contexts_verifications[:i+1]) / (i+1) * contexts_verifications[i] 
                for i in range(len(contexts_verifications))
            ])
        return [num / tot_sum] if tot_sum > 0 else [0.0]
        
    def _evaluate(self, dataset: EvaluationDataset, **kwargs) -> list[float]:
        """Calculate context precision for a single query"""
        contexts_verifications = self._verify_contexts(dataset, **kwargs)
        return self._calculate_context_precision(contexts_verifications)
    
    def _verify_contexts(self, dataset: EvaluationDataset, **kwargs) -> list[int]:
        """Verify if contexts are relevant using LLM"""
        prompts = []
        
        for item in dataset.items:
            prompt = self._build_prompt_with_token_limit(
                template=self._prompt_template,
                contexts=item.contexts,
                question=item.question,
                answer=item.answer
            )
            prompts.append(prompt)

        return self._verify_with_llm(prompts)
    
    def _parse_llm_responses(self, responses: list[BaseMessage]) -> list[int]:
        """Parse LLM responses into context verifications"""
        items = []
        for i, response in enumerate(responses):
            if response is None:
                logger.warning(f"Response {i} is None, defaulting to 0")
                items.append(0)
                continue
                
            try:
                content = response.content.strip()
                items.append(int(content))
            except (json.JSONDecodeError, ValueError):
                fallback_item = _utils.extract_num_from_text(str(response.content))
                if fallback_item is not None:
                    items.append(fallback_item)
                else:
                    logger.warning(f"Failed to parse response {i}: '{response.content}' - defaulting to 0")
                    items.append(0)
            except Exception as e:
                logger.error(f"Error parsing response {i}: {e}")
                items.append(0)
        
        return items


class ContextRecall(RetrievalMetrics):
    """Context Recall: Measures how well retrieval finds ALL relevant information"""
    
    name: str = "context_recall"
    _prompt_template: str = CONTEXT_RECALL_PROMPT
    is_llm_based: bool = True
    
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("name", "context_recall")
        kwargs.setdefault("category", MetricCategory.RETRIEVAL)
        super().__init__(*args, **kwargs)
    
    def _evaluate(self, dataset: EvaluationDataset, **kwargs) -> list[float]:
        """Verify ground truth statements in retrieved contexts"""
        prompts = []
        
        for item in dataset.items:
            combined_text = f"Contexts: {item.contexts}\nGround Truth: {item.ground_truth.contexts}"
            token_count = self._estimate_tokens(combined_text)
            
            if token_count > self.max_prompt_tokens * 0.8:
                available_tokens = int(self.max_prompt_tokens * 0.4)
                
                fitted_contexts = self._fit_contexts_to_token_limit(
                    item.contexts,
                    "placeholder",
                    max_tokens=available_tokens
                )
                
                fitted_ground_truth = self._fit_contexts_to_token_limit(
                    item.ground_truth.contexts,
                    "placeholder", 
                    max_tokens=available_tokens
                )
                
                prompt = self._prompt_template.format(
                    contexts=fitted_contexts,
                    ground_truth=fitted_ground_truth,
                    question=item.question
                )
            else:
                prompt = self._prompt_template.format(
                    contexts=item.contexts,
                    ground_truth=item.ground_truth.contexts,  
                    question=item.question
                )
            
            prompts.append(prompt)
        
        return self._verify_with_llm(prompts)
    
    def _parse_llm_responses(self, responses: list[BaseMessage]) -> list[float]:
        """Parse LLM responses into attribution scores"""
        attributions = []
        for i, response in enumerate(responses):
            if response is None:
                logger.warning(f"Response {i} is None, defaulting to 0.0")
                attributions.append(0.0)
                continue
                
            try:
                data = json.loads(response.content)
                if isinstance(data, list):
                    attributions.extend(data)
                else:
                    attributions.append(float(data))
            except (json.JSONDecodeError, ValueError):
                fallback = _utils.extract_num_from_text(str(response.content))
                if fallback is not None:
                    attributions.append(fallback)
                else:
                    logger.warning(f"Failed to parse LLM response {i}: {response.content}")
                    attributions.append(0.0)
        
        return attributions


# *************************
# Ranking-Based Metrics
# *************************
class MRR(RetrievalMetrics):
    """Mean Reciprocal Rank: Position of first relevant result"""
    
    name: str = "mrr"
    is_llm_based: bool = False
    is_embedding_based: bool = True

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("name", "mrr")
        kwargs.setdefault("category", MetricCategory.RETRIEVAL)
        super().__init__(*args, **kwargs)
    
    def _evaluate(self, dataset: EvaluationDataset, **kwargs) -> list[float]:
        """Calculate MRR with precomputed embeddings (OPTIMIZED)"""
        # PRECOMPUTE ALL EMBEDDINGS FIRST - Major speedup!
        all_contexts = []
        for item in dataset.items:
            all_contexts.append(item.contexts)
            all_contexts.append(item.ground_truth.contexts)
        
        self._precompute_embeddings(all_contexts)
        
        # Now calculate MRR with cached embeddings
        reciprocal_ranks = []
        
        for item in dataset.items:
            rank = self._find_first_relevant_rank(
                retrieved_contexts=item.contexts,
                ground_truth_contexts=item.ground_truth.contexts
            )
            
            rr = 1.0 / rank if rank > 0 else 0.0
            reciprocal_ranks.append(rr)
            
            if rr == 0.0:
                logger.warning(
                    f"MRR: No relevant context found. "
                    f"Retrieved: {len(item.contexts)}, Ground truth: {len(item.ground_truth.contexts)}"
                )
        
        return reciprocal_ranks
    
    def _find_first_relevant_rank(
        self, 
        retrieved_contexts: list[str], 
        ground_truth_contexts: list[str]
    ) -> int:
        """Find rank (1-indexed) of first relevant context"""
        for rank, retrieved_ctx in enumerate(retrieved_contexts, start=1):
            if self._is_match(retrieved_ctx, ground_truth_contexts):
                return rank
        
        logger.debug("No relevant context found in retrieved results")
        return 0


class NDCG(RetrievalMetrics):
    """Normalized Discounted Cumulative Gain: Ranking quality with graded relevance"""
    
    name: str = "ndcg"
    is_llm_based: bool = False
    is_embedding_based: bool = True
    
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("name", "ndcg")
        kwargs.setdefault("category", MetricCategory.RETRIEVAL)
        super().__init__(*args, **kwargs)
    
    def _evaluate(self, dataset: EvaluationDataset, **kwargs) -> list[float]:
        """Calculate NDCG with precomputed embeddings"""
        if not dataset.items:
            return []
        
        # precompute all embeddings first (for speedup) 
        all_contexts = []
        for item in dataset.items:
            all_contexts.append(item.contexts)
            all_contexts.append(item.ground_truth.contexts)
        
        self._precompute_embeddings(all_contexts)
        
        # Calculate NDCG with cached embeddings
        ndcg_scores = []
        for item in dataset.items:
            score = self._calculate_ndcg(item.ground_truth.contexts, item.contexts)
            ndcg_scores.append(score)
            
            if score == 0.0:
                logger.warning(
                    f"NDCG: Zero score. "
                    f"Retrieved: {len(item.contexts)}, Ground truth: {len(item.ground_truth.contexts)}"
                )
        
        return ndcg_scores
    
    def _calculate_ndcg(self, ground_truth: list[str], retrieved: list[str]) -> float:
        """Calculate NDCG for a single query"""
        relevance_map = self._get_relevance_scores(ground_truth, retrieved)
        
        # Calculate DCG (Discounted Cumulative Gain)
        dcg = sum(
            (2 ** relevance_map.get(doc, 0) - 1) / math.log2(i + 2)
            for i, doc in enumerate(retrieved)
        )
        
        # Calculate IDCG (Ideal DCG)
        ideal_scores = sorted(relevance_map.values(), reverse=True)
        idcg = sum(
            (2 ** score - 1) / math.log2(i + 2)
            for i, score in enumerate(ideal_scores)
        )
        
        return dcg / idcg if idcg > 0 else 0.0
    
    def _get_relevance_scores(self, ground_truth: list[str], retrieved: list[str]) -> dict[str, int]:
        """Map retrieved docs to graded relevance scores"""
        relevance_scores = {}
        
        for doc in retrieved:
            rank = self._get_context_rank(doc, ground_truth)
            if rank >= 0:
                relevance_scores[doc] = len(ground_truth) - rank
            else:
                relevance_scores[doc] = 0
        
        return relevance_scores
    
    def _get_context_rank(self, retrieved_context: str, ground_truth_contexts: list[str]) -> int:
        """Find rank (0-indexed) of retrieved context in ground truth"""
        retrieved_clean = retrieved_context.strip()
        
        for idx, gt_ctx in enumerate(ground_truth_contexts):
            gt_clean = gt_ctx.strip()
            
            if retrieved_clean == gt_clean or retrieved_clean.lower() == gt_clean.lower():
                return idx
            
            # Word overlap check
            r_words = set(retrieved_clean.lower().split())
            g_words = set(gt_clean.lower().split())
            overlap = len(r_words & g_words)
            
            if overlap > 0:
                word_overlap = overlap / max(len(r_words | g_words), 1)
                if word_overlap >= 0.6:
                    return idx
        
        # Semantic similarity check with cached embeddings
        if self.embedding_model:
            try:
                retrieved_embedding = self._get_cached_embedding(retrieved_clean)
                
                for idx, gt_ctx in enumerate(ground_truth_contexts):
                    gt_clean = gt_ctx.strip()
                    gt_embedding = self._get_cached_embedding(gt_clean)
                    
                    cosine_sim = self._cosine_similarity(retrieved_embedding, gt_embedding)
                    if cosine_sim >= 0.7:
                        return idx
            except Exception as e:
                logger.warning(f"Embedding-based matching failed: {e}")
        
        return -1



from langchain_core.retrievers import BaseRetriever
from langchain_core.vectorstores import VectorStore
from typing import Any, Optional
from pydantic import PrivateAttr, Field
from typing_extensions import override
from pathlib import Path
from langchain.schema import Document
from loguru import logger
import joblib
import hashlib

# TODO:: adjust if needed and remove redundancy
class Retriever(BaseRetriever):
    """
    Enhanced retrieval with multiple search strategies.
    
    Supports:
    - Semantic search (vector similarity, MMR)
    - Lexical search (BM25, TF-IDF)
    - Hybrid search (combination of semantic + lexical)
    
    Lexical retrievers are automatically saved and loaded from disk.
    """

    _vector_store: VectorStore = PrivateAttr()
    _lexical_retriever: Optional[Any] = PrivateAttr(default=None)
    _lexical_cache_dir: Path = PrivateAttr(default=Path(".gaia_lexical_cache"))
    
    search_type: str = Field(default="similarity", description="Search strategy to use")
    search_kwargs: dict[str, Any] = Field(
        default_factory=dict,
        description="Additional search parameters (k, fetch_k, lambda_mult, score_threshold, etc.)"
    )
    hybrid_weights: dict[str, float] = Field(
        default_factory=lambda: {"semantic": 0.7, "lexical": 0.3},
        description="Weights for hybrid search (must sum to 1.0)"
    )

    def __init__(
        self,
        vector_store: VectorStore,
        search_type: str = "similarity",
        corpus_documents: Optional[list[Document]] = None,
        lexical_cache_dir: Optional[str] = None,
        force_rebuild: bool = False,
        **search_kwargs,
    ):
        """
        Initialize retriever with specified search strategy.
        
        Args:
            vector_store: Vector store for semantic search
            search_type: Type of search strategy (similarity, mmr, bm25, tfidf, hybrid)
            corpus_documents: Documents for lexical search (only needed first time or if force_rebuild=True)
            lexical_cache_dir: Directory for caching lexical retrievers (default: .gaia_lexical_cache)
            force_rebuild: Force rebuilding lexical retriever even if cached version exists
            **search_kwargs: Additional parameters for search (k, max_features, etc.)
        """
        super().__init__()
        self._vector_store = vector_store
        self.search_type = search_type
        self.search_kwargs = search_kwargs
        
        # Set cache directory
        if lexical_cache_dir:
            self._lexical_cache_dir = Path(lexical_cache_dir)
        self._lexical_cache_dir.mkdir(parents=True, exist_ok=True)
        
        # Validate hybrid weights
        if search_type == "hybrid":
            total_weight = sum(self.hybrid_weights.values())
            if not abs(total_weight - 1.0) < 1e-6:
                raise ValueError(f"hybrid_weights must sum to 1.0, got {total_weight}")
        
        # Initialize lexical retriever if needed
        if search_type in ("bm25", "tfidf", "hybrid"):
            self._lexical_retriever = self._get_or_create_lexical_retriever(
                search_type=search_type,
                corpus_documents=corpus_documents,
                force_rebuild=force_rebuild
            )

    def _generate_cache_key(self, search_type: str, corpus_documents: list[Document]) -> str:
        """Generate a unique cache key based on corpus content"""
        # Create hash from document contents
        content_hash = hashlib.md5()
        for doc in corpus_documents:
            content_hash.update(doc.page_content.encode('utf-8'))
        
        # Include search type and number of documents
        cache_key = f"{search_type}_{len(corpus_documents)}_{content_hash.hexdigest()[:16]}"
        return cache_key

    def _get_cache_path(self, search_type: str, corpus_documents: Optional[list[Document]] = None) -> Path:
        """Get the cache file path for the lexical retriever"""
        if corpus_documents:
            cache_key = self._generate_cache_key(search_type, corpus_documents)
        else:
            # Use search type only for loading
            cache_key = search_type
        
        return self._lexical_cache_dir / f"lexical_{cache_key}.pkl"

    def _get_or_create_lexical_retriever(
        self,
        search_type: str,
        corpus_documents: Optional[list[Document]],
        force_rebuild: bool = False
    ) -> Any:
        """Get cached lexical retriever or create a new one"""
        
        # Try to find and load cached retriever
        if not force_rebuild:
            cached_retriever = self._try_load_cached_retriever(search_type)
            if cached_retriever:
                logger.info(f"Loaded cached {search_type} retriever")
                return cached_retriever
        
        # Build new retriever
        if not corpus_documents:
            raise ValueError(
                f"corpus_documents required for {search_type} search. "
                f"No cached retriever found in {self._lexical_cache_dir}"
            )
        
        logger.info(f"Building new {search_type} retriever with {len(corpus_documents)} documents")
        lexical_retriever = self._initialize_lexical_retriever(search_type, corpus_documents)
        
        # Save to cache
        cache_path = self._get_cache_path(search_type, corpus_documents)
        self._save_lexical_retriever(lexical_retriever, cache_path)
        logger.info(f"Cached {search_type} retriever to {cache_path}")
        
        return lexical_retriever

    def _try_load_cached_retriever(self, search_type: str) -> Optional[Any]:
        """Try to load any cached retriever for the given search type"""
        # Look for any cached files matching the search type
        pattern = f"lexical_{search_type}_*.pkl"
        cached_files = list(self._lexical_cache_dir.glob(pattern))
        
        if not cached_files:
            return None
        
        # Use the most recent cache file
        latest_cache = max(cached_files, key=lambda p: p.stat().st_mtime)
        
        try:
            return joblib.load(latest_cache)
        except Exception as e:
            logger.warning(f"Failed to load cached retriever from {latest_cache}: {e}")
            return None

    def _initialize_lexical_retriever(
        self, search_type: str, documents: list[Document]
    ) -> dict[str, Any]:
        """Initialize BM25 or TF-IDF retriever"""
        if search_type in ("bm25", "hybrid"):
            try:
                from rank_bm25 import BM25Okapi
            except ImportError:
                raise ImportError(
                    "rank-bm25 required for BM25 search. "
                    "Install with: pip install rank-bm25"
                )
             
            # Tokenize documents for BM25
            tokenized_docs = [
                doc.page_content.lower().split() for doc in documents
            ]
            return {
                "retriever": BM25Okapi(tokenized_docs),
                "documents": documents,
                "type": "bm25"
            }
        
        elif search_type == "tfidf":
            try:
                from sklearn.feature_extraction.text import TfidfVectorizer
            except ImportError:
                raise ImportError(
                    "scikit-learn required for TF-IDF search. "
                    "Install with: pip install scikit-learn"
                )
            
            # Create TF-IDF vectorizer
            corpus = [doc.page_content for doc in documents]
            vectorizer = TfidfVectorizer(
                lowercase=True,
                stop_words='english',
                max_features=self.search_kwargs.get("max_features", 5000)
            )
            tfidf_matrix = vectorizer.fit_transform(corpus)
            
            return {
                "vectorizer": vectorizer,
                "tfidf_matrix": tfidf_matrix,
                "documents": documents,
                "type": "tfidf"
            }
        
        raise ValueError(f"Unknown lexical search type: {search_type}")

    def _save_lexical_retriever(self, retriever: dict[str, Any], path: Path) -> None:
        """Save lexical retriever to disk"""
        try:
            joblib.dump(retriever, path)
        except Exception as e:
            logger.error(f"Failed to save lexical retriever to {path}: {e}")

    def _retrieve_semantic(self, query: str, k: int = 4) -> list[Document]:
        """Retrieve using vector similarity"""
        search_kwargs = {"k": k}
        
        if self.search_type == "mmr":
            # Add MMR-specific parameters if provided
            if "fetch_k" in self.search_kwargs:
                search_kwargs["fetch_k"] = self.search_kwargs["fetch_k"]
            if "lambda_mult" in self.search_kwargs:
                search_kwargs["lambda_mult"] = self.search_kwargs["lambda_mult"]
            
            return self._vector_store.max_marginal_relevance_search(
                query, **search_kwargs
            )
        else:  # similarity
            return self._vector_store.similarity_search(query, **search_kwargs)

    def _retrieve_lexical(self, query: str, k: int = 4) -> list[Document]:
        """Retrieve using BM25 or TF-IDF"""
        if not self._lexical_retriever:
            raise ValueError("Lexical retriever not initialized")
        
        retriever_type = self._lexical_retriever["type"]
        documents = self._lexical_retriever["documents"]
        
        if retriever_type == "bm25":
            # Tokenize query and get BM25 scores
            tokenized_query = query.lower().split()
            bm25 = self._lexical_retriever["retriever"]
            scores = bm25.get_scores(tokenized_query)
            
            # Get top-k documents
            top_indices = sorted(
                range(len(scores)), key=lambda i: scores[i], reverse=True
            )[:k]
            return [documents[i] for i in top_indices]
        
        elif retriever_type == "tfidf":
            # Transform query and compute similarity
            from sklearn.metrics.pairwise import cosine_similarity
            
            vectorizer = self._lexical_retriever["vectorizer"]
            tfidf_matrix = self._lexical_retriever["tfidf_matrix"]
            
            query_vec = vectorizer.transform([query])
            scores = cosine_similarity(query_vec, tfidf_matrix).flatten()
            
            # Get top-k documents
            top_indices = scores.argsort()[-k:][::-1]
            return [documents[i] for i in top_indices]
        
        return []

    def _retrieve_hybrid(self, query: str, k: int = 4) -> list[Document]:
        """Retrieve using hybrid semantic + lexical search"""
        # Get results from both strategies
        semantic_docs = self._retrieve_semantic(query, k=k * 2)
        lexical_docs = self._retrieve_lexical(query, k=k * 2)
        
        # Combine and re-rank using weighted scores
        doc_scores: dict[str, float] = {}
        
        # Add semantic scores
        sem_weight = self.hybrid_weights.get("semantic", 0.7)
        for idx, doc in enumerate(semantic_docs):
            doc_id = doc.page_content
            # Normalize rank-based score
            score = (len(semantic_docs) - idx) / len(semantic_docs)
            doc_scores[doc_id] = doc_scores.get(doc_id, 0) + score * sem_weight
        
        # Add lexical scores
        lex_weight = self.hybrid_weights.get("lexical", 0.3)
        for idx, doc in enumerate(lexical_docs):
            doc_id = doc.page_content
            score = (len(lexical_docs) - idx) / len(lexical_docs)
            doc_scores[doc_id] = doc_scores.get(doc_id, 0) + score * lex_weight
        
        # Create document map
        doc_map = {doc.page_content: doc for doc in semantic_docs + lexical_docs}
        
        # Sort by combined score and return top-k
        sorted_docs = sorted(
            doc_scores.items(), key=lambda x: x[1], reverse=True
        )
        return [doc_map[doc_id] for doc_id, _ in sorted_docs[:k]]

    def retrieve(self, query: str) -> list[Document]:
        """
        Retrieve documents for query using configured search strategy.
        
        Args:
            query: Search query string
            
        Returns:
            List of relevant documents
        """
        k = self.search_kwargs.get("k", 4)
        
        if self.search_type in ("similarity", "mmr"):
            return self._retrieve_semantic(query, k)
        elif self.search_type in ("bm25", "tfidf"):
            return self._retrieve_lexical(query, k)
        elif self.search_type == "hybrid":
            return self._retrieve_hybrid(query, k)
        
        raise ValueError(f"Unknown search type: {self.search_type}")
 
    @override
    def _get_relevant_documents(
        self,
        query: str,
        **kwargs
    ) -> list[Document]:
        """LangChain retriever interface implementation"""
        return self.retrieve(query)
    
    def clear_cache(self) -> None:
        """Clear all cached lexical retrievers"""
        if self._lexical_cache_dir.exists():
            for cache_file in self._lexical_cache_dir.glob("lexical_*.pkl"):
                cache_file.unlink()
                logger.info(f"Deleted cache file: {cache_file}")


from rag_opt.dataset import EvaluationDatasetItem, GroundTruth, ComponentUsage, TrainDataset, EvaluationDataset
from langchain.agents import AgentExecutor, create_tool_calling_agent
from langchain.tools.retriever import create_retriever_tool
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnablePassthrough
from langchain_community.vectorstores import VectorStore
from typing_extensions import Annotated, Doc, Optional
from concurrent.futures import Future, Executor, as_completed
from rag_opt._utils import get_shared_executor
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnableSequence
from rag_opt.rag.callbacks import RAGCallbackHandler
from rag_opt.rag.retriever import Retriever
from rag_opt.rag.reranker import BaseReranker
from langchain.schema import Document
from rag_opt.llm import RAGLLM
from loguru import logger
from enum import Enum
import time


class RAGMode(Enum):
    """Enumeration of supported RAG execution modes."""
    BASELINE = "baseline"
    AGENTIC = "agentic"
    HYBRID = "hybrid"
    RERANKING = "reranking"


class BaseRAGInterface:
    """
    Base interface for custom RAG implementations.
    Users can extend this class to create custom RAG workflows.
    """
    
    def get_answer(self, query: str, **kwargs) -> EvaluationDatasetItem:
        """Process a single query and return evaluation item."""
        raise NotImplementedError("Subclasses must implement get_answer()")
    
    def get_batch_answers(self, dataset: TrainDataset, **kwargs) -> EvaluationDataset:
        """Process multiple queries in batch."""
        raise NotImplementedError("Subclasses must implement get_batch_answers()")
    
    def get_relevant_docs(self, query: str, **kwargs) -> list[Document]:
        """Retrieve relevant documents for a query."""
        raise NotImplementedError("Subclasses must implement get_relevant_docs()")


class RAGWorkflow(BaseRAGInterface):
    """
    Production-ready RAG pipeline with multiple execution modes.
    
    Supports:
    - Baseline: Standard retrieval + generation
    - Agentic: Tool-calling agents with decision-making
    - Hybrid: Semantic + lexical search combination
    - Reranking: Document reranking for improved relevance
    """
    
    def __init__(
        self, 
        embeddings, 
        vector_store: VectorStore, 
        llm: Annotated[RAGLLM, Doc("LLM for answer generation")],
        reranker: Optional[BaseReranker] = None,
        retrieval_config: Optional[dict] = None,
        *,
        corpus_documents: Annotated[Optional[list[Document]], Doc("Corpus documents for lexical search")] = None,
        lexical_cache_dir: Annotated[Optional[str], Doc("Path to lexical retriever cache")] = None,
        embedding_provider_name: Annotated[Optional[str], Doc("Embedding provider name")] = None,
        embedding_model_name: Annotated[Optional[str], Doc("Embedding model name")] = None,
        llm_provider_name: Annotated[Optional[str], Doc("LLM provider name")] = None,
        llm_model_name: Annotated[Optional[str], Doc("LLM model name")] = None,
        reranker_provider_name: Annotated[Optional[str], Doc("Reranker provider name")] = None,
        reranker_model_name: Annotated[Optional[str], Doc("Reranker model name")] = None,
        vector_store_provider_name: Annotated[Optional[str], Doc("Vector store provider name")] = None,
        max_workers: Annotated[int, Doc("Maximum workers for parallel processing")] = 5,
        executor: Annotated[Optional[Executor], Doc("Thread pool executor")] = None,
        **kwargs
    ):
        self.embeddings = embeddings
        self.vector_store = vector_store
        self.llm = llm
        self.reranker = reranker

        # Provider and model metadata for tracking
        self.embedding_provider_name = embedding_provider_name
        self.embedding_model_name = embedding_model_name
        self.llm_provider_name = llm_provider_name
        self.llm_model_name = llm_model_name
        self.reranker_provider_name = reranker_provider_name
        self.reranker_model_name = reranker_model_name
        self.vector_store_provider_name = vector_store_provider_name

        # Initialize retrieval configuration
        retrieval_config = retrieval_config or {
            "search_type": kwargs.get("search_type", "similarity"), 
            "k": kwargs.get("k", 5)
        }
        
        # Initialize retriever with caching support for hybrid mode
        self.retrieval = Retriever(
            vector_store, 
            corpus_documents=corpus_documents,
            lexical_cache_dir=lexical_cache_dir,
            **retrieval_config
        )
        
        # Executor for parallel batch processing
        self.executor = executor or get_shared_executor(max_workers)
        
        # Create baseline RAG chain
        self.rag_chain = self._create_rag_chain()
        
        # Initialize agentic components
        self.agent_executor: Optional[AgentExecutor] = None
        self._init_agent()

    def _init_agent(self) -> None:
        """Initialize agent executor for agentic mode."""
        retrieval_tool = create_retriever_tool(
            self.retrieval,
            "retrieve_relevant_context",
            "Search and return information required to answer the question accurately",
        )
        
        agent_prompt = ChatPromptTemplate.from_messages([
            ("system", "You are a helpful assistant. Use the retrieve_relevant_context tool to get relevant information to answer questions accurately."),
            ("human", "{input}"),
            ("placeholder", "{agent_scratchpad}"),
        ])
        
        agent = create_tool_calling_agent(self.llm, [retrieval_tool], agent_prompt)
        self.agent_executor = AgentExecutor(
            agent=agent,
            tools=[retrieval_tool],
            verbose=False,
        )

    def _create_rag_prompt(self) -> ChatPromptTemplate:
        """Create RAG prompt template for baseline mode."""
        template = """Answer the question based only on the following context:

{context}

Question: {question}

Answer:"""
        return ChatPromptTemplate.from_template(template)
    
    def _format_docs(self, docs: list[Document]) -> str:
        """Format documents into a single context string."""
        return "\n\n".join(doc.page_content for doc in docs)
    
    def _create_rag_chain(self) -> RunnableSequence:
        """Create the baseline RAG chain."""
        prompt = self._create_rag_prompt()
        
        rag_chain = (
            {"context": self.retrieval | self._format_docs, "question": RunnablePassthrough()}
            | prompt
            | self.llm
            | StrOutputParser()
        )
        return rag_chain

    def _generate_eval_metadata(self, callback_handler: RAGCallbackHandler) -> dict[str, ComponentUsage]:
        """Generate evaluation metadata for metrics tracking."""
        return {
            "cost": ComponentUsage(
                llm=callback_handler.llm_stats.total_cost,
                embedding=callback_handler.embedding_stats.total_cost,
                vectorstore=0.0,
                reranker=callback_handler.reranker_stats.total_cost
            ),
            "latency": ComponentUsage(
                llm=callback_handler.llm_stats.total_latency,
                embedding=callback_handler.embedding_stats.total_latency,
                vectorstore=0.0,
                reranker=callback_handler.reranker_stats.total_latency
            )
        }

    def _apply_reranking(
        self,
        query: str,
        docs: list[Document],
        rerank_top_k: int,
        callback_handler: RAGCallbackHandler
    ) -> list[Document]:
        """Apply reranking to retrieved documents."""
        if not self.reranker:
            logger.warning("Reranker not initialized, returning original documents")
            return docs
        
        start_time = time.time()
        reranked_docs = self.reranker.rerank(
            query=query, 
            documents=docs, 
            top_k=rerank_top_k
        )
        rerank_latency = time.time() - start_time
        
        callback_handler.track_reranking(docs, rerank_latency)
        
        return reranked_docs

    def _get_callback_handler(self, verbose: bool = False) -> RAGCallbackHandler:
        """Create a callback handler with current configuration."""
        return RAGCallbackHandler(
            verbose=verbose,
            llm_provider_name=self.llm_provider_name,
            llm_model_name=self.llm_model_name,
            embedding_provider_name=self.embedding_provider_name,
            embedding_model_name=self.embedding_model_name,
            reranker_model_name=self.reranker_model_name,
            reranker_provider_name=self.reranker_provider_name,
            vector_store_provider_name=self.vector_store_provider_name
        )

    def get_answer(
        self, 
        query: str,
        *,
        ground_truth: Optional[GroundTruth] = None,
        mode: RAGMode = RAGMode.BASELINE,
        use_reranker: bool = False,
        rerank_top_k: int = 10,
        verbose: bool = False,
        **kwargs
    ) -> EvaluationDatasetItem:
        """
        Process query through RAG pipeline with configurable execution mode.
        
        Args:
            query: Question to answer
            ground_truth: Optional ground truth for evaluation
            mode: Execution mode (BASELINE, AGENTIC, HYBRID, RERANKING)
            use_reranker: Whether to apply reranking (overrides mode if True)
            rerank_top_k: Number of documents to keep after reranking
            verbose: Enable verbose logging
            
        Returns:
            EvaluationDatasetItem with answer and metadata
        """
        # Determine effective mode
        if use_reranker:
            mode = RAGMode.RERANKING
        
        callback_handler = self._get_callback_handler(verbose)
        
        # Route to appropriate mode handler
        if mode == RAGMode.AGENTIC:
            return self._get_agentic_answer(query, ground_truth, callback_handler)
        elif mode in (RAGMode.RERANKING, RAGMode.HYBRID):
            return self._get_reranked_answer(query, ground_truth, rerank_top_k, callback_handler)
        else:  # BASELINE
            return self._get_baseline_answer(query, ground_truth, callback_handler)

    def _get_baseline_answer(
        self,
        query: str,
        ground_truth: Optional[GroundTruth],
        callback_handler: RAGCallbackHandler
    ) -> EvaluationDatasetItem:
        """Execute baseline RAG without reranking."""
        response = self.rag_chain.invoke(
            query, 
            config={"callbacks": [callback_handler]}
        )
        contexts = callback_handler.retrieved_contexts
        metadata = self._generate_eval_metadata(callback_handler)
        
        return EvaluationDatasetItem(
            question=query,
            answer=response,
            contexts=contexts,
            ground_truth=ground_truth,
            metadata=metadata
        )

    def _get_reranked_answer(
        self,
        query: str,
        ground_truth: Optional[GroundTruth],
        rerank_top_k: int,
        callback_handler: RAGCallbackHandler
    ) -> EvaluationDatasetItem:
        """Execute RAG with document reranking."""
        # Retrieve documents
        docs = self.retrieval.invoke(query)
        
        # Apply reranking
        reranked_docs = self._apply_reranking(query, docs, rerank_top_k, callback_handler)
        
        # Create temporary chain with reranked documents
        prompt = self._create_rag_prompt()
        temp_chain = (
            {"context": lambda _: self._format_docs(reranked_docs), "question": RunnablePassthrough()}
            | prompt
            | self.llm
            | StrOutputParser()
        )
        
        response = temp_chain.invoke(query, config={"callbacks": [callback_handler]})
        contexts = [doc.page_content for doc in reranked_docs]
        metadata = self._generate_eval_metadata(callback_handler)
        
        return EvaluationDatasetItem(
            question=query,
            answer=response,
            contexts=contexts,
            ground_truth=ground_truth,
            metadata=metadata
        )

    def _get_agentic_answer(
        self,
        query: str,
        ground_truth: Optional[GroundTruth],
        callback_handler: RAGCallbackHandler
    ) -> EvaluationDatasetItem:
        """Execute agentic RAG with tool-calling agent."""
        if not self.agent_executor:
            raise RuntimeError("Agent executor not initialized")
        
        response = self.agent_executor.invoke(
            {"input": query}, 
            config={"callbacks": [callback_handler]}
        )
        
        contexts = callback_handler.retrieved_contexts
        metadata = self._generate_eval_metadata(callback_handler)
        
        return EvaluationDatasetItem(
            question=query,
            answer=response.get("output"),
            contexts=contexts,
            ground_truth=ground_truth,
            metadata=metadata
        )

    def get_batch_answers(
        self, 
        dataset: TrainDataset,
        mode: RAGMode = RAGMode.BASELINE,
        use_reranker: bool = False,
        rerank_top_k: int = 10,
        verbose: bool = False,
        **kwargs
    ) -> EvaluationDataset:
        """
        Get answers for all dataset questions in parallel.
        
        Args:
            dataset: Training dataset with questions
            mode: Execution mode for all queries
            use_reranker: Whether to apply reranking
            rerank_top_k: Number of documents to keep after reranking
            verbose: Enable verbose logging
            
        Returns:
            EvaluationDataset with all answers
        """
        futures: list[Future[EvaluationDatasetItem]] = []

        # Submit all tasks in parallel
        for item in dataset.items:
            future = self.executor.submit(
                self.get_answer, 
                query=item.question, 
                ground_truth=item.to_ground_truth(),
                mode=mode,
                use_reranker=use_reranker,
                rerank_top_k=rerank_top_k,
                verbose=verbose,
                **kwargs
            )
            futures.append(future)
        
        # Collect results as they complete
        items: list[EvaluationDatasetItem] = []
        for future in as_completed(futures):
            try:
                items.append(future.result(timeout=120)) 
            except Exception as e:
                logger.error(f"Error processing question: {e}")
        
        return EvaluationDataset(items=items)
    
    def get_relevant_docs(
        self, 
        query: str,
        use_reranker: bool = False,
        rerank_top_k: int = 10
    ) -> list[Document]:
        """
        Retrieve relevant documents for query with optional reranking.
        
        Args:
            query: Search query
            use_reranker: Whether to apply reranking
            rerank_top_k: Number of documents to keep after reranking
            
        Returns:
            List of relevant documents
        """
        docs = self.retrieval.invoke(query)
        
        if use_reranker and self.reranker:
            callback_handler = self._get_callback_handler(verbose=False)
            docs = self._apply_reranking(query, docs, rerank_top_k, callback_handler)
        
        return docs

    def store_documents(self, documents: list[Document]) -> None:
        """Store documents in vector store."""
        try:
            self.vector_store.add_documents(documents)
            logger.info(f"Successfully stored {len(documents)} documents")
        except Exception as e:
            logger.error(f"Failed to store documents: {e}")
            raise

    def clear_cache(self) -> None:
        """Clear lexical retriever cache if applicable."""
        if hasattr(self.retrieval, 'clear_cache'):
            self.retrieval.clear_cache()
            logger.info("Retrieval cache cleared")



from langchain_core.vectorstores import VectorStore
from langchain_core.documents import Document
from loguru import logger 


class Indexer:
    """ Take embeddings and store in vector db / index (Phase 3) """

    def __init__(self, 
                 chunk_size:int,
                 chunk_overlap:int,
                 vector_store:VectorStore,
                 ) -> None:
        """ Load the parsed documents and split them

            Args: (those args represents hyperparameters in our RAG pipeline)
                - chunk_size: Size of each chunk.
                - chunk_overlap: Overlap between chunks.
                - vector_store: Vector store to use.
                    https://python.langchain.com/docs/integrations/vectorstores/

        """
        self.vector_store = vector_store
        self.embeddings = vector_store.embeddings
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

    def store(self, documents: list[Document]):
        """Store documents in vector store"""
        if not documents:
            raise ValueError("No documents provided to store")
        
        self.vector_store.add_documents(documents)
        logger.success(f"Successfully stored {len(documents)} documents in vector store")
    



def index_documents(documents: list[Document], 
                    chunk_size:int, 
                    chunk_overlap:int,
                    vector_store:VectorStore,
                    ):
    """ Index documents in vector store """
    indexer = Indexer(chunk_size, chunk_overlap, vector_store)
    indexer.store(documents)
    logger.success(f"Successfully indexed {len(documents)} documents in vector store")



def index_data_files(dataset_path: str="./data", 
                    **kwargs,):
    from rag_opt.rag.parser import Parser
    parser = Parser(path=dataset_path, **kwargs)
    docs = parser.load_docs() 
    index_documents(docs, **kwargs)